/***************************************************************************
                          arrayitem.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Aug 28 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QFile>
#include <QPainter>
#include <QPointF>
#include <QRegExp>
#include <QStringList>
#include <QTextStream>
#include <QVector>
#include <KConfigGroup>
#include <KMessageBox>
#include "arrayitem.h"
#include "frameitem.h"
#include "kplscene.h"
#include "utils.h"

/**
  * Determine number of columns of an array to be read from a text stream.
  * @param t text stream to read from.
  * @param idec decimal character: Point or Comma.
  * @return number of columns, 0 for error.
  */
static int nColumns(QTextStream& t,
                    ArrayItem::DecimalCharacter idec = ArrayItem::Point);

int nColumns(QTextStream& t, ArrayItem::DecimalCharacter idec)
{
  int nc = 0;
  while (!t.atEnd()) {
    QString line = t.readLine().trimmed();
    if ((!line.isEmpty()) && (line[0] != '#')) {
      nc = line.split(QRegExp(idec ? "[ \t]" : "[ ,\t]"),
                      QString::SkipEmptyParts).count();
      break;
    }
  }
  return nc;
}

ArrayItem::ArrayItem() :
 m_internal(false), m_errorBars(false), m_colX(0), m_colY(0), m_colErr(0),
 m_rowStart(0), m_count(0), m_numRows(0), m_numCols(0),
 m_decimalCharacter(Point), m_data(0)
{
}

ArrayItem::ArrayItem(unsigned col, int colX, int colY, int colErr,
                     bool errorBars, Symbol symbol,
                     DecimalCharacter decimalChar, const QUrl& u, double xn,
                     double yn, QGraphicsItem* parent, int np, int rowStart,
                     FillStyle fill, double xOffset, double yOffset,
                     bool visible, bool intern, double relSize) :
 ScaledItem(fill, symbol, col, xn, yn, xOffset, yOffset, relSize, parent),
 m_internal(intern), m_errorBars(errorBars), m_colX(colX), m_colY(colY),
 m_colErr(colErr), m_rowStart(0), m_count(0), m_numRows(0), m_numCols(0),
 m_decimalCharacter(decimalChar), m_data(0), m_url(u)
{
  if (!visible)
    hide();
  if (m_internal) {
    m_count = m_numRows = qMax(np, 0);
    m_numCols = 3;
    m_decimalCharacter = Point;
    int dim = m_numCols * m_numRows;
    m_data = new double*[m_numCols];
    m_data[0] = new double[dim];
    memset(m_data[0], 0, dim * sizeof(double));
    for (int j = 1; j < m_numCols; j++)
      m_data[j] = m_data[j - 1] + m_numRows;
  } else
    if (!m_url.url().isEmpty()) {
      if (readFile()) {
        setRange(rowStart, (np < 0) ? m_numRows : np);
        adjustCols();
      } else
        KMessageBox::error(0, m_url.url());
    }
}

ArrayItem::ArrayItem(const ArrayItem& a) :
 ScaledItem(a), m_internal(a.m_internal), m_errorBars(a.m_errorBars),
 m_colX(a.m_colX), m_colY(a.m_colY), m_colErr(a.m_colErr),
 m_rowStart(a.m_rowStart), m_count(a.m_count), m_numRows(a.m_numRows),
 m_numCols(a.m_numCols), m_decimalCharacter(a.m_decimalCharacter),
 m_data(ArrayItem::copyArray(a.m_data, a.m_numCols, a.m_numRows)),
 m_url(a.m_url)
{
}

ArrayItem::ArrayItem(const KConfigGroup* plo, const QUrl& urlPlo,
                     unsigned color, int colX, int colY, int colErr,
                     bool errorBars, Symbol symbol, QGraphicsItem* parent) :
 ScaledItem(plo, symbol, color, parent), m_errorBars(false), m_colX(0),
 m_colY(0), m_colErr(0), m_rowStart(0), m_count(0), m_data(0)
{
  m_internal = plo->readEntry("internal", false);
  if (m_internal) {
    m_count = m_numRows = plo->readEntry("n", 0);
    m_errorBars = plo->readEntry("errbars", 0);
    m_numCols = 3;
    m_decimalCharacter = Point;
    m_data = new double*[m_numCols];
    m_data[0] = new double[m_numCols * m_numRows];
    for (int j = 1; j < m_numCols; j++)
      m_data[j] = m_data[j - 1] + m_numRows;
    for (int i = 0; i < m_numRows; i++) {
      QStringList list = plo->readEntry(QString("r%1").arg(i)).split(' ');
      int cnt = list.count();
      for (int j = 0; j < m_numCols; j++)
        m_data[j][i] = (j < cnt) ? list[j].toDouble() : 0;
    }
  } else {
    m_url = absoluteUrl(plo->readEntry("path", ""), urlPlo);
    m_decimalCharacter = DecimalCharacter(plo->readEntry("idec", 0));
    if (!readFile()) {
      KMessageBox::error(0, m_url.url());
      return;
    }
    setRange(plo->readEntry("istart", 0), plo->readEntry("n", m_numRows));
    m_errorBars = plo->readEntry("errbars", errorBars);
  }
  adjustCols(plo->readEntry("ix", colX), plo->readEntry("iy", colY),
             plo->readEntry("ie", colErr));
}

ArrayItem::~ArrayItem()
{
  freeArray(&m_data);
}

void ArrayItem::adjustCols(int colX, int colY, int colErr)
{
  int icolmax = m_numCols - 1;
  m_colX = qMin(colX, icolmax);
  m_colY = qMin(colY, icolmax);
  m_colErr = qMin(colErr, icolmax);
}

QRectF ArrayItem::boundingRect() const
{
  if (isVisible() && bRect.isNull())
    paintCond(0);
  return bRect;
}

double** ArrayItem::copyArray(double** source, int nCols, int nRows)
{
  double** x2 = 0;
  if (source) {
    int n = nCols * nRows;
    x2 = new double*[nCols];
    if (x2) {
      x2[0] = 0;
      x2[0] = new double[n];
      if (x2[0]) {
        memcpy(x2[0], source[0], n * sizeof(double));
        for (int i = 1; i < nCols; i++)
          x2[i] = x2[i - 1] + nRows;
      }
    }
  }
  return x2;
}

void ArrayItem::exponent(int& expX, int& expY, double& fx, double& fy) const
{
  if (m_data) {
    double xmin, xmax, ymin, ymax;
    limits(xmin, xmax, ymin, ymax);
    ScaledItem::exponent(qMax(fabs(xmin), fabs(xmax)), expX, fx);
    ScaledItem::exponent(qMax(fabs(ymin), fabs(ymax)), expY, fy);
  }
}

void ArrayItem::freeArray(double*** data)
{
  if (*data) {
    delete [] *data[0];
    delete [] *data;
    *data = 0;
  }
}

void ArrayItem::limits(double& xMin, double& xMax, double& yMin,
                       double& yMax) const
{
  if (m_data) {
    ScaledItem::limits(xMin, xMax, &m_data[m_colX][m_rowStart], m_count);
    ScaledItem::limits(yMin, yMax, &m_data[m_colY][m_rowStart], m_count);
    xMin += m_xOffset;
    xMax += m_xOffset;
    yMin += m_yOffset;
    yMax += m_yOffset;
  }
}

void ArrayItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*,
                      QWidget*)
{
  paintCond(painter);
}

int ArrayItem::readFile()
{
  m_numRows = readFile(m_url, m_numCols, &m_data, m_decimalCharacter);
  return m_numRows;
}

int ArrayItem::readFile(const QUrl& url, int& nCols, double*** data,
                        DecimalCharacter decimalCharacter)
{
  freeArray(data);
  int nRows = 0;
  nCols = 0;
  QString fileName;
  if (url.isLocalFile())
    fileName = url.path();
  else
    if (!Utils::download(url, fileName, 0))
      return nRows;
  QFile f(fileName);
  if (f.open(QIODevice::ReadOnly))
    nRows = readFile(&f, nCols, data, decimalCharacter);
  if (!url.isLocalFile())
    QFile::remove(fileName);
  return nRows;
}

int ArrayItem::readFile(QIODevice* device, int& nCols, double*** data,
                        DecimalCharacter decimalCharacter)
{
  {
    QTextStream t(device);
    nCols = nColumns(t);
  }
  int nRows = 0;
  if (nCols) {
    device->reset();
    QTextStream t(device);
    nRows = readLines(t, nCols, data, decimalCharacter);
    device->close();
  }
  return nRows;
}

int ArrayItem::readFile(QString* text, int& nCols, double*** data,
                        DecimalCharacter decimalCharacter)
{
  {
    QTextStream t(text, QIODevice::ReadOnly);
    nCols = nColumns(t);
  }
  int nRows = 0;
  if (nCols) {
    QTextStream t(text, QIODevice::ReadOnly);
    nRows = readLines(t, nCols, data, decimalCharacter);
  }
  return nRows;
}

int ArrayItem::readFile(QString* text)
{
  m_numRows = readFile(text, m_numCols, &m_data);
  return m_numRows;
}

int ArrayItem::readLines(QTextStream& t, int nCols, double*** data,
                         DecimalCharacter decimalChar)
{
  QVector<double> x2;
  int np = 0;
  int k = 0;
  int nr = 0;
  while (!t.atEnd()) {
    QString line = t.readLine().trimmed();
    if ((!line.isEmpty()) && (line[0] != '#')) {
      ArrayItem::translate(line, decimalChar);
      QStringList list = line.split(QRegExp(decimalChar ? "[ \t]" : "[ ,\t]"),
                                    QString::SkipEmptyParts);
      nr++;
      np += nCols;
      x2.resize(np);
      int cnt = list.count();
      for (int i = 0; i < nCols; i++) {
        x2[k] = (i < cnt) ? list[i].toDouble() : 0.0;
        k++;
      }
    }
  }
  ArrayItem::freeArray(data);
  *data = new double*[nCols];
  (*data)[0] = new double[np];
  for (int i = 1; i < nCols; i++)
    (*data)[i] = (*data)[i - 1] + nr;
  k = 0;
  for (int j = 0; j < nr; j++)
    for (int i = 0; i < nCols; i++)
      (*data)[i][j] = x2[k++];
  return nr;
}

void ArrayItem::setDashOffset()
{
  if (!m_fillStyle  && (m_symbol > 1)) {
    myDashOffset = m_dashOffset;
    m_pen.setDashOffset(m_dashOffset);
    double s = 0;
    foreach (double d, m_pen.dashPattern())
      s += d;
    m_dashOffset += length(&m_data[m_colX][m_rowStart],
                           &m_data[m_colY][m_rowStart],
                           m_count) / m_pen.widthF();
    m_dashOffset -= int(m_dashOffset / s) * s;
  }
}

void ArrayItem::setData(double** data)
{
  freeArray(&m_data);
  m_data = data;
};

void ArrayItem::setRange(int rowStart, int n)
{
  m_rowStart = qMin(rowStart, m_numRows - 1);
  m_count = qMin(n, m_numRows - m_rowStart);
}

QPainterPath ArrayItem::shape() const
{
  if (isVisible() && bPath.isEmpty())
    paintCond(0);
  return bPath;
}

void ArrayItem::translate(QString& line, DecimalCharacter decimalChar)
{
  if (decimalChar)
    line.replace(',', '.');
  line.replace('D', 'E');
  line.replace('d', 'e');
}

void ArrayItem::writePlo(KConfigGroup* plo, const QUrl& urlPlo, bool abs) const
{
  plo->writeEntry("Type", "ARRAYITEM");
  ScaledItem::writePlo(plo, isVisible());
  plo->writeEntry("internal", m_internal);
  plo->writeEntry("ix", m_colX);
  plo->writeEntry("iy", m_colY);
  plo->writeEntry("ie", m_colErr);
  plo->writeEntry("errbars", m_errorBars ? 1 : 0);
  plo->writeEntry("n", m_count);
  if (m_internal) {
    plo->writeEntry("istart", 0);
    for (int i = 0; i < m_count; i++) {
      QStringList list;
      for (int j = 0; j < m_numCols; j++)
        list.append(QString::number(m_data[j][i + m_rowStart]));
      plo->writeEntry(QString("r%1").arg(i), list.join(" "));
    }
  } else {
    plo->writeEntry("idec", int(m_decimalCharacter));
    plo->writeEntry("istart", m_rowStart);
    plo->writeEntry("path", path(urlPlo, m_url, abs));
  }
}

void ArrayItem::paintCond(QPainter* painter) const
{
  if (!m_data)
    return;
  FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
  KplScene* sc = 0;
  bool useFrame = frame && frame->isVisible();
  if (!useFrame)
    sc = static_cast<KplScene*>(scene());
  if (painter)
    painter->setClipRect(useFrame ? frame->clipRect() : sc->sceneRect());
  if (m_symbol != -17) {
    if (painter) {
      if (m_fillStyle) {
        painter->setBrush(createBrush(m_color, m_fillStyle));
        painter->setPen(Qt::NoPen);
      } else {
        m_pen.setDashOffset(myDashOffset);
        painter->setPen(m_pen);
      }
    }
    plotArray(painter, &m_data[m_colX][m_rowStart],
              &m_data[m_colY][m_rowStart], m_count);
  }
  if (m_errorBars) {
    if (painter) {
      QPen pen = m_pen;
      pen.setStyle(Qt::SolidLine);
      painter->setPen(pen);
    }
    int iend = m_count + m_rowStart;
    QPointF dx = QPointF(0.5 * symbRect.right(), 0);
    for (int i = m_rowStart; i < iend; ++i) {
      double xa = m_normX * (m_xOffset + m_data[m_colX][i]);
      double ya = m_normY * (m_yOffset + m_data[m_colY][i]);
      if (sc ? (FrameItem::valid(xa) && FrameItem::valid(ya)) :
               frame->valid(xa, ya)) {
        QPointF pu(xa, ya);
        QPointF p = (sc ? sc->map(pu) : frame->mapFromUser(pu));
        QPointF ph(xa, ya + m_normY * m_data[m_colErr][i]);
        QPointF dy = p - (sc ? sc->map(ph) : frame->mapFromUser(ph));
        if (dy.y() >= symbRect.bottom()) {
          QPointF dy0 = QPointF(0, symbRect.bottom());
          QPointF tl = -dx - dy;
          if (painter) {
            painter->save();
            painter->translate(p);
            painter->drawLine(QLineF(tl, dx - dy));
            painter->drawLine(QLineF(-dx + dy, -tl));
            painter->drawLine(QLineF(-dy0, -dy));
            painter->drawLine(QLineF(dy0, dy));
            painter->restore();
          } else {
            double d = 0.5 * m_pen.widthF();
            QRectF r = QRectF(tl, - tl).adjusted(-d, -d, d, d).translated(p);
            if (bRect.isNull())
              bRect = r;
            else
              bRect |= r;
          }
        }
      }
    }
  }
}
