/***************************************************************************
                          frameitem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FRAMEITEM_H
#define FRAMEITEM_H

#include <math.h>
#include <QGraphicsRectItem>
#include "textitem.h"

/**
  * Frame item class. Class for frame items with tics, grid, labels.
  * @author Werner Stille
  */
class Q_DECL_EXPORT FrameItem : public KplItem, public QGraphicsRectItem
{
public:
  /** Grid mode. Determines how the frame should be plotted. */
  enum GridMode {
    FrameOnly,      ///< Only rectangular frame.
    Axes,           ///< Axes with tics and without labels.
    AxesWithLabels, ///< Axes with tics and labels.
    Grid,           ///< Grid without labels.
    GridWithLabels  ///< Grid with labels.
  };
  /** Specific QGraphicsItem type number for function type() */
  enum {
    Type = UserType + Frame
    /**< Specific QGraphicsItem type number of FrameItem for function type(). */
  };
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x0 left page margin in cm.
    * @param y0 bottom page margin in cm.
    * @param width width in cm.
    * @param height height in cm.
    * @param colorFrame color for frame and labels.
    * @param colorGrid color for grids and tics.
    * @param logX true for logarithmic x axis.
    * @param logY true for logarithmic y axis.
    * @param gridMode grid mode.
    * @param xMax x value of right data window margin in user units.
    * @param yMax y value of top window data margin in user units.
    * @param textX x axis text.
    * @param textY y axis text.
    * @param header heading text.
    * @param relSize relative size of text, markers and lines.
    * @param numDigX number of digits for x axis.
    * @param numDigY number of digits for y axis.
    * @param minTicX number of intervals between major tics of x axis.
    * @param minTicY number of intervals between major tics of y axis.
    * @param expX logarithm of normalization factor for x axis.
    * @param expY logarithm of normalization factor for y axis.
    * @param xMin x value of left data window margin in user units.
    * @param yMin y value of bottom data window margin in user units.
    * @param xTic distance between major tics of x axis.
    * @param yTic distance between major tics of y axis.
    * @param collapsed true for collapsed in list view.
    * @param visible true for visible item.
    * @param xOffset offset for first major x tick.
    * @param yOffset offset for first major y tick.
    * @param xBottom true for x axis labels on bottom side.
    * @param xTop true for x axis labels on top side.
    * @param yLeft true for y axis labels on left side.
    * @param yRight true for y axis labels on right side.
    */
  FrameItem(double x0 = 4, double y0 = 3, double width = 15,
            double height = 10,  unsigned colorFrame = 0,
            unsigned colorGrid = 0, bool logX = false, bool logY = false,
            GridMode gridMode = AxesWithLabels, double xMax = 15,
            double yMax = 10, const QString& textX = "",
            const QString& textY = "", const QString& header = "",
            double relSize = 1, int numDigX = -1, int numDigY = -1,
            int minTicX = 5, int minTicY = 2, int expX = 0, int expY = 0,
            double xMin = 0, double yMin = 0, double xTic = 5, double yTic = 2,
            bool collapsed = false, bool visible = true, double xOffset = 0,
            double yOffset = 0, bool xBottom = true, bool xTop = false,
            bool yLeft = true, bool yRight = false);
  /**
    * Copy constructor.
    * @param a source item.
    */
  FrameItem(const FrameItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param x0 default value for left page margin in cm.
    * @param y0 default value for bottom page margin in cm.
    * @param width default value for width in cm.
    * @param height default value for height in cm.
    * @param colorFrame default value for color for frame and labels.
    * @param colorGrid default value for color for grids and tics.
    * @param gridMode grid mode.
    */
  FrameItem(const KConfigGroup* plo, double x0 = 4, double y0 = 3,
            double width = 15, double height = 10, unsigned colorFrame = 0,
            unsigned colorGrid = 0, GridMode gridMode = AxesWithLabels);
  /** Destructor. */
  virtual ~FrameItem();
  /**
   * Calculates scaling parameters.
   * @param xMin pointer to double for rounded minimum value.
   * @param xMax pointer to double for rounded maximum value.
   * @param tick pointer to double for major tick distance.
   * @param minTic pointer to int for intervals between major tics.
   * @param numDig pointer to int for number of places after decimal point
   * for lettering.
   * @param xmi minimum value.
   * @param xma maximum value.
   * @param f normalization factor.
   * @param c scaling constant, desired value for tick / (xma - xmi).
   * @param log true for logarithmic axis.
   */
  static void autoScale(double& xMin, double& xMax, double& tick, int& minTic,
                        int& numDig, double xmi, double xma, double f,
                        double c, bool log);
  /**
   * Calculates and sets the scaling parameters.
   * @param xmi minimum x value.
   * @param xma maximum x value.
   * @param fx x normalization factor.
   * @param ymi minimum y value.
   * @param yma maximum y value.
   * @param fy y normalization factor.
   */
  void autoScale(double xmi, double xma, double fx, double ymi, double yma,
                 double fy);
  /**
   * Calculates and sets the scaling parameters.
   * @param autoNorm true for automatic normalization.
   */
  void autoScale(bool autoNorm = true);
  /** Calculates bounding rectangle. */
  virtual QRectF boundingRect() const;
  /**
   * Calculates scaling parameters for a frame item and all its child items.
   * @param autoNorm true for automatic normalization.
   * @param items list of child items.
   * @param fx normalization factor for x axis.
   * @param fy normalization factor for y axis.
   * @return true when scaling was successful.
   */
  bool calcAutoScale(bool autoNorm, const QList<QGraphicsItem*>& items,
                     double& fx, double& fy);
  /** Returns clipping rectangle. */
  inline QRectF clipRect() const { return cRect; };
  /** Returns color for frame and labels. */
  inline unsigned colorFrame() const { return m_colorFrame; };
  /** Returns color for grids and tics. */
  inline unsigned colorGrid() const { return m_colorGrid; };
  /** Returns logarithm of normalization factor for x axis. */
  inline int expX() const { return m_expX; };
  /** Returns logarithm of normalization factor for y axis. */
  inline int expY() const { return m_expY; };
  /** Returns grid mode. */
  inline GridMode gridMode() const { return m_gridMode; };
  /** Returns heading text. */
  inline QString header() const { return m_header; };
  /** Returns height in cm. */
  inline double height() const { return m_height; };
  /** Returns collapse state for list view. */
  inline bool collapsed() const { return m_collapsed; };
  /** Returns true for logarithmic x axis. */
  inline bool logX() const { return m_logX; };
  /** Returns true for logarithmic y axis. */
  inline bool logY() const { return m_logY; };
  /**
   * Maps from user to pixel coordinates.
   * @param p user coordinates.
   * @return pixel coordinates.
   */
  QPointF mapFromUser(const QPointF& p) const;
  /**
   * Maps from pixel to user coordinates.
   * @param p pixel coordinates.
   * @return user coordinates.
   */
  QPointF mapToUser(const QPointF& p) const;
  /**
   * Maps from pixel to user coordinates.
   * @param rect rectangle given in pixel coordinates.
   * @return rectangle given in user coordinates.
   */
  QRectF mapToUser(const QRectF& rect) const;
  /** Returns number of intervals between major tics of x axis. */
  inline int minTicX() const { return m_minTicX; };
  /** Returns number of intervals between major tics of y axis. */
  inline int minTicY() const { return m_minTicY; };
  /** Returns normalization factor for x values. */
  inline double normX() const { return m_normX; };
  /** Returns normalization factor for y values. */
  inline double normY() const { return m_normY; };
  /** Returns number of digits for x axis. */
  inline int numDigX() const { return m_numDigX; };
  /** Returns number of digits for y axis. */
  inline int numDigY() const { return m_numDigY; };
  /** Paints item. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /** Returns position in cm units. */
  QPointF rectPos() const;
  /**
    * Sets color for frame and labels.
    * @param color color.
    */
  inline void setColorFrame(unsigned color) { m_colorFrame = color; };
  /**
    * Sets color for grids and tics.
    * @param color color.
    */
  inline void setColorGrid(unsigned color) { m_colorGrid = color; };
  /**
    * Sets exponent of x normalization factor.
    * @param expX exponent.
    */
  inline void setExpX(int expX) { m_expX = expX; m_normX = pow(10, expX); };
  /**
    * Sets exponent of y normalization factor.
    * @param expY exponent.
    */
  inline void setExpY(int expY) { m_expY = expY; m_normY = pow(10, expY); };
  /**
    * Sets grid mode.
    * @param gridMode grid mode.
    */
  inline void setGridMode(GridMode gridMode) { m_gridMode = gridMode; };
  /**
    * Sets height.
    * @param height height in cm.
    */
  inline void setHeight(double height) { m_height = height; };
  /**
    * Sets logarithmic mode of x axis
    * @param logX true for logarithmic x axis.
    */
  inline void setLogX(bool logX) { m_logX = logX; };
  /**
    * Sets logarithmic mode of y axis
    * @param logY true for logarithmic y axis.
    */
  inline void setLogY(bool logY) { m_logY = logY; };
  /**
    * Sets number of intervals between major tics of x axis.
    * @param minTicX number of intervals.
    */
  inline void setMinTicX(int minTicX) { m_minTicX = minTicX; };
  /**
    * Sets number of intervals between major tics of y axis.
    * @param minTicY number of intervals.
    */
  inline void setMinTicY(int minTicY) { m_minTicY = minTicY; };
  /**
    * Sets number of digits for x axis.
    * @param numDigX number of digits.
    */
  inline void setNumDigX(int numDigX) { m_numDigX = numDigX; };
  /**
    * Sets number of digits for y axis.
    * @param numDigY number of digits.
    */
  inline void setNumDigY(int numDigY) { m_numDigY = numDigY; };
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /**
    * Sets heading text.
    * @param text heading text.
    */
  inline void setHeader(const QString& text) { m_header = text; };
  /**
    * Sets x axis text.
    * @param text axis text.
    */
  inline void setTextX(const QString& text) { m_textX = text; };
  /**
    * Sets y axis text.
    * @param text axis text.
    */
  inline void setTextY(const QString& text) { m_textY = text; };
  /**
    * Sets width.
    * @param width width in cm.
    */
  inline void setWidth(double width) { m_width = width; };
  /**
    * Sets left page margin.
    * @param x0 page margin in cm.
    */
  inline void setX0(double x0) { m_x0 = x0; };
  /**
    * Sets bottom side label mode.
    * @param xBottom true for x axis labels on bottom side.
    */
  inline void setXBottom(bool xBottom) { m_xBottom = xBottom; };
  /**
    * Sets upper x data limit.
    * @param xMax x value of right data window margin in user units.
    */
  inline void setXMax(double xMax) { m_xMax = xMax; };
  /**
    * Sets lower x data limit.
    * @param xMin x value of left data window margin in user units.
    */
  inline void setXMin(double xMin) { m_xMin = xMin; };
  /**
    * Sets x offset.
    * @param xOffset offset for first major x tick.
    */
  inline void setXOffset(double xOffset) { m_xOffset = xOffset; };
  /**
    * Sets x tick distance.
    * @param xTic distance between major tics of x axis.
    */
  inline void setXTic(double xTic) { m_xTic = xTic; };
  /**
    * Sets top side label mode.
    * @param xTop true for x axis labels on top side.
    */
  inline void setXTop(bool xTop) { m_xTop = xTop; };
  /**
    * Sets bottom page margin.
    * @param y0 page margin in cm.
    */
  inline void setY0(double y0) { m_y0 = y0; };
  /**
    * Sets left side label mode.
    * @param yLeft true for x axis labels on left side.
    */
  inline void setYLeft(bool yLeft) { m_yLeft = yLeft; };
  /**
    * Sets upper y data limit.
    * @param yMax y value of upper data window margin in user units.
    */
  inline void setYMax(double yMax) { m_yMax = yMax; };
  /**
    * Sets lower y data limit.
    * @param yMin y value of lower data window margin in user units.
    */
  inline void setYMin(double yMin) { m_yMin = yMin; };
  /**
    * Sets y offset.
    * @param yOffset offset for first major y tick.
    */
  inline void setYOffset(double yOffset) { m_yOffset = yOffset; };
  /**
    * Sets right side label mode.
    * @param yRight true for x axis labels on right side.
    */
  inline void setYRight(bool yRight) { m_yRight = yRight; };
  /**
    * Sets y tick distance.
    * @param yTic distance between major tics of y axis.
    */
  inline void setYTic(double yTic) { m_yTic = yTic; };
  /** Returns user units steps for one pixel steps. */
  QPointF steps() const;
  /** Returns x axis text. */
  inline QString textX() const { return m_textX; };
  /** Returns y axis text. */
  inline QString textY() const { return m_textY; };
  /** Returns specific QGraphicsItem type number. */
  virtual int type() const { return Type; };
  /** Toggles collapse state. */
  inline void toggleCollapsed() { m_collapsed = !m_collapsed; };
  /** Applies changes and forces redrawing. */
  void update();
  /**
   * Sets position.
   * @param p position of bottom left corner in cm.
   */
  void updatePos(QPointF p);
  /**
   * Checks a double value for validity.
   * @param v double value.
   * @param log true for use as argument of log function.
   * @return true for valid.
   */
  static inline bool valid(double v, bool log = false)
    { return (v == v) && (fabs(v) <= 1e300) && ((!log) || (v > 0)); };
  /**
   * Checks coordinates for validity.
   * @param x x value.
   * @param y y value.
   * @return true for valid.
   */
  bool valid(double x, double y) const;
  /** Returns width in cm. */
  inline double width() const { return m_width; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   */
  void writePlo(KConfigGroup* plo) const;
  /** Returns left page margin in cm. */
  inline double x0() const { return m_x0; };
  /** Returns true for x axis labels on bottom side. */
  inline bool xBottom() const { return m_xBottom; };
  /** Returns x value of right data window margin in user units. */
  inline double xMax() const { return m_xMax; };
  /** Returns x value of left data window margin in user units. */
  inline double xMin() const { return m_xMin; };
  /** Returns offset for first major x tick. */
  inline double xOffset() const { return m_xOffset; };
  /** Returns distance between major tics of x axis. */
  inline double xTic() const { return m_xTic; };
  /** Returns true for x axis labels on top side. */
  inline bool xTop() const { return m_xTop; };
  inline double y0() const { return m_y0; };
  /** Returns true for y axis labels on left side. */
  inline bool yLeft() const { return m_yLeft; };
  /** Returns true for y axis labels on right side. */
  inline bool yRight() const { return m_yRight; };
  /** Returns y value of upper data window margin in user units. */
  inline double yMax() const { return m_yMax; };
  /** Returns y value of lower data window margin in user units. */
  inline double yMin() const { return m_yMin; };
  /** Returns offset for first major y tick. */
  inline double yOffset() const { return m_yOffset; };
  /** Returns distance between major tics of y axis. */
  inline double yTic() const { return m_yTic; };

protected:
  /**
   * Plots text.
   * @param painter pointer to QPainter object.
   * @param text text.
   * @param x x position in pixel units.
   * @param y y position in pixel units.
   * @param alignment horizontal alignment.
   * @param dir text direction in degree.
   */
  void drawText(QPainter* painter, const QString& text, double x, double y,
                TextItem::Alignment alignment, double dir = 0) const;
  /**
   * Increment logarithm for next major axis tick.
   * @param x logarithm of tick value.
   * @param f factor to get second tick in a decade.
   * @return mantissa of x.
   */
  double incLog(double& x, double f) const;
  /** Updates position. */
  virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent* event);
  /**
   * Prepares number label.
   * @param v value in user units.
   * @param numDig number of digits after decimal point
   * (-1: no display of decimal point).
   * @param label label (output).
   * @return length of label string.
   */
  int number(double v, int numDig, QString& label) const;
  /**
   * Paints or calculates bounding area.
   * @param painter QPainter object, 0 for calculating bounding area.
   */
  void paintCond(QPainter* painter) const;
  /** Sets properties according to transformation matrix. */
  void setProperties();
  /**
   * Calculates actual distance between major tics.
   * @param tic tick distance in user units.
   * @param log true for logarithmic axis.
   * @return optimized value in case of logarithmic axis.
   */
  double tick(double tic, bool log) const;
  bool m_logX, m_logY, m_xBottom, m_xTop, m_yLeft, m_yRight, m_collapsed;
  int m_numDigX, m_numDigY, m_minTicX, m_minTicY, m_expX, m_expY;
  GridMode m_gridMode;
  unsigned m_colorFrame, m_colorGrid;
  double m_x0, m_y0, m_width, m_height, m_xMin, m_xMax, m_yMin, m_yMax, m_xTic,
    m_yTic, m_xOffset, m_yOffset, m_normX, m_normY, xminl, xmaxl, yminl, ymaxl;
  QString m_textX, m_textY, m_header;
  mutable QRectF bRect;
  QRectF cRect;
  QTransform md, m_inv, m_matrix;
};

#endif
