//
// C++ Implementation: kplscene
//
// Description: 
//
//
// Author: Werner Stille <kpl@werner-stille.de>, (C) 2017
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include <QFileInfo>
#include <QGraphicsSceneMouseEvent>
#include <QGraphicsView>
#include <QHash>
#include <QPrinter>
#include <QProcess>
#include <QSvgGenerator>
#include <QTextStream>
#include <KConfig>
#include <KConfigGroup>
#define TRANSLATION_DOMAIN "libkpl"
#include <KLocalizedString>
#include <KMessageBox>
// #include <KMimeTypeTrader>
// #include <KToolInvocation>
#include <kmimetypetrader.h>
#include <ktoolinvocation.h>
#include "arcitem.h"
#include "array3ditem.h"
#include "arrowitem.h"
#include "fun3ditem.h"
#include "ellipseitem.h"
#include "imageitem.h"
#include "kplscene.h"
#include "legenditem.h"
#include "lineitem.h"
#include "parfunitem.h"
#include "psgraph.h"
#include "rectitem.h"
#include "scalebaritem.h"
#include "splineitem.h"
#include "textitem.h"
#include "utils.h"

#include <QDebug>

static QHash<QString, int> itemTypes;
static bool initialized = false;

static bool checkList(const QList<QGraphicsItem*>& list, Qt::SortOrder order)
{
  int z = -1;
  foreach (QGraphicsItem* item, list) {
    int z2 = item->zValue();
    if ((item == list.first()) ||
        ((order == Qt::DescendingOrder) && (z2 < z)) ||
        ((order == Qt::AscendingOrder) && (z2 > z)))
      z = z2;
    else {
//       qDebug() << "checkList(): not sorted: z1 = " << z << ", z2 = " << z2;
      qDebug("checkList(): not sorted");
      return false;
    }
  }
  return true;
}

static bool greaterThan(QGraphicsItem* it1, QGraphicsItem* it2)
{
  return it1->zValue() > it2->zValue();
}

static bool lessThan(QGraphicsItem* it1, QGraphicsItem* it2)
{
  return it1->zValue() < it2->zValue();
}

KplScene::KplScene(const QSizeF& size, QObject* parent) :
 QGraphicsScene(parent)
{
// TODO: check problems with BSP and 3D items or complex item lists.
//   setItemIndexMethod(NoIndex);
  setBspTreeDepth(2);
  QWidget w;
  m_sx = w.logicalDpiX() / 2.54;
  m_sy = -w.logicalDpiY() / 2.54;
  setSize(size);
}

KplScene::~KplScene()
{
}

bool KplScene::exportImage(const QUrl& url, bool boundingBox)
{
  QString path = url.path();
  QString suffix = QFileInfo(path).suffix();
  QString imageFile = url.isLocalFile() ? path :
                                          Utils::tempFileName("." + suffix);
  bool success = false;
  if (suffix == "svg") {
// TODO: QSvgGenerator still does not support clipping.
    QSvgGenerator generator;
    generator.setFileName(imageFile);
    QRectF rect = boundingBox ? itemsBoundingRect() : sceneRect();
    QSizeF sz = rect.size();
    QRectF dest = QRectF(QPointF(), sz);
    generator.setSize(sz.toSize());
    generator.setViewBox(dest);
    generator.setResolution(m_sx * 2.54);
    generator.setTitle(url.fileName());
    QPainter painter(&generator);
    clearSelection();
    render(&painter, dest, rect);
    success = true;
  } else {
    QString opt;
    if (suffix == "pdf")
      opt = "pdfwrite";
    else if (suffix == "png")
      opt = "png16m";
    else if (suffix == "ppm")
      opt = suffix;
    else if (suffix == "bmp")
      opt = "bmp16m";
    else if ((suffix == "jpg") || (suffix == "jpeg"))
      opt = "jpeg";
    else if ((suffix == "tif") || (suffix == "tiff"))
      opt = "tiff24nc";
    if (opt.isEmpty())
      return false;
    opt += (suffix == "pdf") ?
      " -dPDFSETTINGS=/prepress -dEmbedAllFonts=true" :
      " -r600 -dGraphicsAlphaBits=4 -dTextAlphaBits=4";
    QString epsFile = Utils::tempFileName(".eps");
    if ((success = savePS(Utils::url(epsFile), false, boundingBox))) {
      QProcess p;
      p.start(QString("gs -sOutputFile=%1 -q -dSAFER -dBATCH -dNOPAUSE "
                      "-dEPSCrop -sDEVICE=%2 %3").arg(imageFile).arg(opt).
              arg(epsFile));
      success = p.waitForFinished();
       QFile::remove(epsFile);
    }
  }
  if (success && (!url.isLocalFile()))
    Utils::move(imageFile, url);
  return success;
}

bool KplScene::framelike(int type)
{
  return (type == FrameItem::Type) || (type == Array3DItem::Type) ||
         (type == Fun3DItem::Type);
}

void KplScene::insertItem(QGraphicsItem* item, int z)
{
  int z1 = z;
  if (z < 0) {
    QList<QGraphicsItem*> list = items(Qt::DescendingOrder);
    z1 = (list.count()) ? (list.first()->zValue() + 1) : 0;
  }
  item->setZValue(z1);
  if (framelike(item->type()))
    addItem(item);
  else {
    FrameItem* fr = 0;
    if ((fr = frame(z)))
      item->setParentItem(fr);
    else
      addItem(item);
  }
  setProperties(item);
}

QList<QGraphicsItem*> KplScene::items(Qt::SortOrder order) const
{
  QList<QGraphicsItem*> list = QGraphicsScene::items(order);
  if (!checkList(list, order)) {
    qDebug("KplScene::items()");
    qSort(list.begin(), list.end(), order ? greaterThan : lessThan);
  }
  return list;
}

QPointF KplScene::map(const QPointF& p) const
{
  return m_matrix.map(p);
}

QPointF KplScene::mapInv(const QPointF& p) const
{
  return m_inv.map(p);
}

void KplScene::print(QPrinter* printer)
{
  printer->setFullPage(true);
  QPainter painter(printer);
  painter.setRenderHint(QPainter::Antialiasing);
  clearSelection();
  QSizeF sz = printer->resolution() * m_size / 2.54;
  render(&painter,
         QRectF(QPointF(0, printer->paperRect().height() - sz.height()), sz),
         sceneRect(), Qt::IgnoreAspectRatio);
}

bool KplScene::readData(const QUrl& url,
                        ArrayItem::DecimalCharacter decimalChar,
                        bool autoPathHeader, bool autoNorm, bool logX,
                        bool logY, double x0, double y0, double width,
                        double height, unsigned colorFrame, unsigned colorGrid,
                        unsigned colorData, FrameItem::GridMode gridMode,
                        const QString& xText, const QString& yText,
                        const QString& header, double relSize, int colX,
                        int colY, int colErr, bool errorBars,
                        KplItem::Symbol symbol)
{
  FrameItem* fi = new FrameItem(x0, y0, width, height, colorFrame, colorGrid,
                                logX, logY, gridMode, width, height, xText,
                                yText, header, relSize);
  insertItem(fi);
  ArrayItem* ai = new ArrayItem(colorData, colX, colY, colErr, errorBars,
                                symbol, decimalChar, url, 1, 1, 0, -1);
  bool success = ai->numRows();
  if (success) {
    ai->setErrorBars(errorBars && (colErr < ai->numCols()));
    insertItem(ai);
    fi->autoScale(autoNorm);
    if (autoPathHeader && (ai->zValue() == 1))
      fi->setHeader(url.isLocalFile() ? url.path() : url.url());
  } else {
    delete ai;
    delete fi;
  }
  return success;
}

bool KplScene::readPlo(const QUrl& url, QWidget* window)
{
  QList<QGraphicsItem*> list;
  bool success = readPlo(url, list, window);
  if (success)
    foreach (QGraphicsItem* item, list)
      insertItem(item);
  return success;
}

bool KplScene::readPlo(const QUrl& url, QList<QGraphicsItem*>& list,
                       QWidget* window, unsigned colorFrame,
                       unsigned colorGrid, unsigned colorData,
                       KplItem::Symbol type, double x0, double y0,
                       double width, double height,
                       FrameItem::GridMode gridMode, int colX, int colY, int colErr,
                       bool errorBars)
{
  if (!initialized) {
    itemTypes["FRAMEITEM"] = KplItem::Frame;
    itemTypes["ARRAYITEM"] = KplItem::Array;
    itemTypes["FUNITEM"] = KplItem::Function;
    itemTypes["PARFUNITEM"] = KplItem::ParFunction;
    itemTypes["SPLINEITEM"] = KplItem::Spline;
    itemTypes["ARRAY3DITEM"] = KplItem::Array3D;
    itemTypes["FUN3DITEM"] = KplItem::Function3D;
    itemTypes["LEGENDITEM"] = KplItem::Legend;
    itemTypes["TEXTITEM"] = KplItem::Text;
    itemTypes["LINEITEM"] = KplItem::Line;
    itemTypes["ARROWITEM"] = KplItem::Arrow;
    itemTypes["ARCITEM"] = KplItem::Arc;
    itemTypes["RECTITEM"] = KplItem::Rectangle;
    itemTypes["ELLIPSEITEM"] = KplItem::Ellipse;
    itemTypes["IMAGEITEM"] = KplItem::Image;
    itemTypes["SCALEBARITEM"] = KplItem::ScaleBar;
    initialized = true;
  }
  QString tmpfile;
  bool success = Utils::download(url, tmpfile, window);
  if (success) {
    QFileInfo f2(tmpfile);
    success = f2.isReadable();
    if (success) {
      KConfig cfg(tmpfile, KConfig::SimpleConfig);
      KConfigGroup plo = cfg.group("Items");
      if (int nItems = qMax(plo.readEntry("nitems", 0), 0)) {
        for (int i = 0; i < nItems; ++i) {
          plo = cfg.group(QString("Item %1").arg(i));
          switch (itemTypes.value(plo.readEntry("Type"))) {
            case KplItem::Arc:
              list << new ArcItem(&plo, colorFrame);
              break;
            case KplItem::Array:
              list << new ArrayItem(&plo, url, colorData, colX, colY, colErr,
                                    errorBars, type);
              break;
            case KplItem::Array3D:
              list << new Array3DItem(&plo, url, x0, y0, width, height,
                                      colorFrame, colorGrid, colorData,
                                      gridMode, colX, colY, colErr, errorBars);
              break;
            case KplItem::Arrow:
              list << new ArrowItem(&plo, colorFrame);
              break;
            case KplItem::Ellipse:
              list << new EllipseItem(&plo, colorFrame);
              break;
            case KplItem::Frame:
              list << new FrameItem(&plo, x0, y0, width, height, colorFrame,
                                    colorGrid, gridMode);
              break;
            case KplItem::Function:
              list << new FunItem(&plo, url, colorData, type);
              break;
            case KplItem::Function3D:
              list << new Fun3DItem(&plo, url, x0, y0, width, height,
                                    colorFrame, colorGrid, colorData,
                                    gridMode);
              break;
            case KplItem::Image:
              list << new ImageItem(&plo, url);
              break; 
            case KplItem::Legend:
              list << new LegendItem(&plo, colorData, colorFrame, type, width);
              break;
            case KplItem::Line:
              list << new LineItem(&plo, colorFrame);
              break;
            case KplItem::ParFunction:
              list << new ParFunItem(&plo, url, colorData, type);
              break;
            case KplItem::Rectangle:
              list << new RectItem(&plo, colorFrame);
              break;
            case KplItem::ScaleBar:
              list << new ScaleBarItem(&plo, colorFrame);
              break;
            case KplItem::Spline:
              list << new SplineItem(&plo, colorData, type);
              break;
            case KplItem::Text:
              list << new TextItem(&plo, colorFrame);
            }
          }
        } else
          success = false;
      }
      if (!url.isLocalFile())
        QFile::remove(tmpfile);
    }
  return success;
}

bool KplScene::savePlo(const QUrl& url, bool absolute, char frm,
                       int precision) const
{
  return savePlo(url, items(Qt::AscendingOrder), absolute, frm, precision);
}

bool KplScene::savePlo(const QUrl& url, const QList<QGraphicsItem*>& list,
                       bool absolute, char frm, int precision)
{
  QString filename = url.isLocalFile() ? url.path() : Utils::tempFileName();
  bool success = true;
  if (QFile::exists(filename))
    success = QFile::remove(filename);
  if (success) {
    KConfig cfg(filename, KConfig::SimpleConfig);
    KConfigGroup plo = cfg.group("Items");
    int n = list.count();
    plo.writeEntry("nitems", n);
    for (int i = 0; i < n; i++) {
      KConfigGroup plo = cfg.group(QString("Item %1").arg(i));
      QGraphicsItem* item = list.at(i);
      switch (item->type()) {
        case ArcItem::Type:
          qgraphicsitem_cast<ArcItem*>(item)->writePlo(&plo);
          break;
        case ArrayItem::Type:
          qgraphicsitem_cast<ArrayItem*>(item)->writePlo(&plo, url, absolute);
          break;
        case Array3DItem::Type:
          qgraphicsitem_cast<Array3DItem*>(item)->writePlo(&plo, url, absolute);
          break;
        case ArrowItem::Type:
          qgraphicsitem_cast<ArrowItem*>(item)->writePlo(&plo);
          break;
        case EllipseItem::Type:
          qgraphicsitem_cast<EllipseItem*>(item)->writePlo(&plo);
          break;
        case FrameItem::Type:
          qgraphicsitem_cast<FrameItem*>(item)->writePlo(&plo);
          break;
        case FunItem::Type:
          qgraphicsitem_cast<FunItem*>(item)->writePlo(&plo, url, absolute,
                                                       frm, precision);
          break;
        case Fun3DItem::Type:
          qgraphicsitem_cast<Fun3DItem*>(item)->writePlo(&plo, url, absolute,
                                                         frm, precision);
          break;
        case ImageItem::Type:
          qgraphicsitem_cast<ImageItem*>(item)->writePlo(&plo, url, absolute);
          break;
        case LegendItem::Type:
          qgraphicsitem_cast<LegendItem*>(item)->writePlo(&plo);
          break;
        case LineItem::Type:
          qgraphicsitem_cast<LineItem*>(item)->writePlo(&plo);
          break;
        case ParFunItem::Type:
          qgraphicsitem_cast<ParFunItem*>(item)->writePlo(&plo, url, absolute,
                                                          frm, precision);
          break;
        case RectItem::Type:
          qgraphicsitem_cast<RectItem*>(item)->writePlo(&plo);
          break;
        case ScaleBarItem::Type:
          qgraphicsitem_cast<ScaleBarItem*>(item)->writePlo(&plo);
          break;
        case SplineItem::Type:
          qgraphicsitem_cast<SplineItem*>(item)->writePlo(&plo, frm,
                                                          precision);
          break;
        case TextItem::Type:
          qgraphicsitem_cast<TextItem*>(item)->writePlo(&plo);
      }
    }
  }
  if (success && (!url.isLocalFile()))
    success = Utils::move(filename, url);
  return success;
}

bool KplScene::savePS(const QUrl& url, bool landscape, bool boundingBox)
{
  if (!QGraphicsScene::items().count())
    return false;
  bool eps = (QFileInfo(url.path()).suffix() == "eps");
  QString suffix = eps ? ".eps" : ".ps";
  QString filename = (url.isEmpty() || (!url.isLocalFile())) ?
    Utils::tempFileName(suffix) : url.path();
  QString tmpFile = boundingBox ? Utils::tempFileName(suffix) : filename;
  QSizeF sz = size();
  if (landscape)
    sz.transpose();
  bool success = true;
  QFile f(tmpFile);
  if ((success = f.open(QIODevice::WriteOnly))) {
    QTextStream ts(&f);
    PSGraph g(&ts, items(Qt::AscendingOrder), size(), landscape, eps);
  }
  if (boundingBox) {
    QProcess p;
    p.start(QString("gs -q -dSAFER -dBATCH -dNOPAUSE "
                    "-sDEVICE=bbox %1").arg(tmpFile));
    if ((success = p.waitForFinished(-1))) {
      QByteArray bbox = p.readAllStandardError();
      QFile f1(tmpFile);
      f1.open(QIODevice::ReadOnly);
      QFile f2(filename);
      f2.open(QIODevice::WriteOnly);
      bool find = true;
      while (!f1.atEnd()) {
        QByteArray line = f1.readLine();
        if (find)
          if (line.indexOf("%%BoundingBox:") >= 0) {
            line = bbox;
            find = false;
          }
        f2.write(line);
      }
      f1.remove();
    } else
      KMessageBox::error(0, i18n("Bounding box calculation failed!"));
  }
  if (url.isEmpty()) {
    emit previewCreated(filename);
// TODO: startServiceByDesktopName deprecated
    KToolInvocation::startServiceByDesktopName(KMimeTypeTrader::self()->
      preferredService("application/postscript")->desktopEntryName(),
      filename);
  } else
  if (!url.isLocalFile())
    Utils::move(filename, url);
  return true;
}

void KplScene::setDashOffsets() const
{
  KplItem::resetDashOffset();
  foreach (QGraphicsItem* item, items(Qt::AscendingOrder))
    if (item->isVisible())
      switch (item->type()) {
        case ArcItem::Type: {
          ArcItem* it = qgraphicsitem_cast<ArcItem*>(item);
          it->setDashOffset();
          it->update();
          break;
        }
        case ArrayItem::Type: {
          ArrayItem* it = qgraphicsitem_cast<ArrayItem*>(item);
          it->setDashOffset();
          it->update();
          break;
        }
        case FunItem::Type: {
          FunItem* it = qgraphicsitem_cast<FunItem*>(item);
          it->setDashOffset();
          it->update();
          break;
        }
        case LegendItem::Type:
          KplItem::resetDashOffset();
          break;
        case LineItem::Type: {
          LineItem* it = qgraphicsitem_cast<LineItem*>(item);
          it->setDashOffset();
          it->update();
          break;
        }
        case ParFunItem::Type: {
          ParFunItem* it = qgraphicsitem_cast<ParFunItem*>(item);
          it->setDashOffset();
          it->update();
          break;
        }
        case SplineItem::Type: {
          SplineItem* it = qgraphicsitem_cast<SplineItem*>(item);
          it->setDashOffset();
          it->update();
        }
      }
}

void KplScene::setProperties(QGraphicsItem* item) const
{
  switch (item->type()) {
    case ArcItem::Type:
      qgraphicsitem_cast<ArcItem*>(item)->setProperties(m_matrix);
      break;
    case ArrayItem::Type:
      qgraphicsitem_cast<ArrayItem*>(item)->setProperties(m_matrix);
      break;
    case Array3DItem::Type:
      qgraphicsitem_cast<Array3DItem*>(item)->setProperties(m_matrix);
      break;
    case ArrowItem::Type:
      qgraphicsitem_cast<ArrowItem*>(item)->setProperties(m_matrix);
      break;
    case EllipseItem::Type:
      qgraphicsitem_cast<EllipseItem*>(item)->setProperties(m_matrix);
      break;
    case FrameItem::Type:
      qgraphicsitem_cast<FrameItem*>(item)->setProperties(m_matrix);
      break;
    case FunItem::Type:
      qgraphicsitem_cast<FunItem*>(item)->setProperties(m_matrix);
      break;
    case Fun3DItem::Type:
      qgraphicsitem_cast<Fun3DItem*>(item)->setProperties(m_matrix);
      break;
    case ImageItem::Type:
      qgraphicsitem_cast<ImageItem*>(item)->setProperties(m_matrix);
      break;
    case LegendItem::Type:
      qgraphicsitem_cast<LegendItem*>(item)->setProperties(m_matrix);
      break;
    case LineItem::Type:
      qgraphicsitem_cast<LineItem*>(item)->setProperties(m_matrix);
      break;
    case RectItem::Type:
      qgraphicsitem_cast<RectItem*>(item)->setProperties(m_matrix);
      break;
    case ParFunItem::Type:
      qgraphicsitem_cast<ParFunItem*>(item)->setProperties(m_matrix);
      break;
    case ScaleBarItem::Type:
      qgraphicsitem_cast<ScaleBarItem*>(item)->setProperties(m_matrix);
      break;
    case SplineItem::Type:
      qgraphicsitem_cast<SplineItem*>(item)->setProperties(m_matrix);
      break;
    case TextItem::Type:
      qgraphicsitem_cast<TextItem*>(item)->setProperties(m_matrix);
  }
}

void KplScene::setSize(const QSizeF& size)
{
  m_size = size;
  m_matrix = QTransform().scale(m_sx, m_sy).translate(0, -m_size.height());
  m_inv = QTransform(1 / m_sx, 0, 0, 1 / m_sy, 0, m_size.height());
  setSceneRect(m_matrix.mapRect(QRectF(QPointF(), m_size)));
  foreach (QGraphicsItem* item, items(Qt::AscendingOrder))
    setProperties(item);
  foreach (QGraphicsView* view, views())
    view->centerOn(map(QPointF()));
}

QPointF KplScene::steps(int z) const
{
  FrameItem* fr = frame(z);
  return fr ? fr->steps() : QPointF(1 / m_sx, -1 / m_sy);
}

void KplScene::updatePos(QGraphicsItem* item)
{
  emit positionChanged(item);
}

FrameItem* KplScene::frame(int z) const
{
  FrameItem* fr = 0;
  if (z)
    foreach (QGraphicsItem* item, items(Qt::DescendingOrder))
      if (framelike(item->type()) && item->isVisible() &&
          ((z < 0) || (item->zValue() < z))) {
        switch (item->type()) {
          case FrameItem::Type:
            fr = qgraphicsitem_cast<FrameItem*>(item);
            break;
          case Array3DItem::Type:
            fr = qgraphicsitem_cast<Array3DItem*>(item);
            break;
          case Fun3DItem::Type:
            fr = qgraphicsitem_cast<Fun3DItem*>(item);
        }
        break;
      }
  return fr;
}

void KplScene::mouseMoveEvent(QGraphicsSceneMouseEvent* e)
{
  QGraphicsScene::mouseMoveEvent(e);
  QPointF p = e->scenePos();
  const FrameItem* fr = 0;
  if (e->buttons() == Qt::LeftButton) {
    QGraphicsItem* item = mouseGrabberItem();
    if (item) {
      p = item->pos();
      switch (item->type()) {
        case FrameItem::Type:
          p += map(qgraphicsitem_cast<FrameItem*>(item)->rectPos());
          break;
        case Array3DItem::Type:
          p += map(qgraphicsitem_cast<Array3DItem*>(item)->rectPos());
          break;
        case Fun3DItem::Type:
          p += map(qgraphicsitem_cast<Fun3DItem*>(item)->rectPos());
          break;
      }
    }
  } else
    foreach (QGraphicsItem* item,
             QGraphicsScene::items(p, Qt::IntersectsItemShape,
             Qt::DescendingOrder))
      if (item->type() == FrameItem::Type) {
        fr = qgraphicsitem_cast<FrameItem*>(item);
        break;
      }
  emit mouseMoved(p, fr);
}

void KplScene::mousePressEvent(QGraphicsSceneMouseEvent* e)
{
  QGraphicsScene::mousePressEvent(e);
  emit mousePressed(e->scenePos(), e->buttons());
}
