//
// C++ Interface: kplscene
//
// Description: 
//
//
// Author: Werner Stille <tille@uni-freiburg.de>, (C) 2017
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef KPLSCENE_H
#define KPLSCENE_H

#include <QSizeF>
#include <QGraphicsScene>
#include <QUrl>
#include "arrayitem.h"
#include "frameitem.h"

class QPointF;
class QPrinter;
class QString;
class QTransform;
class FrameItem;

/**
  * Kpl graphics scene class.
  * The methods #readData, #readPlo and #insertItem can be used to populate the
  * scene with plot items. The methods #print and #savePS allow to print and
  * to export as PostScript file, respectively.
  * @author Werner Stille
  */
class Q_DECL_EXPORT KplScene : public QGraphicsScene
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param size scene size in cm units.
   * @param parent pointer to parent QObject.
   */
  KplScene(const QSizeF& size = QSizeF(21, 14.8), QObject* parent = 0);
  /** Destructor. */
  ~KplScene();
  /**
   * Exports as image file.
   * @param url URL of the file.
   * @param boundingBox true for calculation of bounding box.
   * @return true when file is saved.
   */
  bool exportImage(const QUrl& url, bool boundingBox = true);
  /**
   * Checks if type of an item is that of a FrameItem or an item inheriting
   * from FrameItem.
   * @param type item type as returned by the item's function type()
   * reimplemented from QGraphicsItem::type().
   * @return true for type equal to #FrameItem::Type, Array3DItem::Type or
   * Fun3DItem::Type.
   */
  static bool framelike(int type);
  /**
   * Inserts a plot item and makes it a child item of a #framelike item, if a
   * framelike item with lower z value exists and the inserted item is no
   * framelike item itself.
   * @param item pointer to plot item.
   * @param z z value of item, -1 for appending.
   */
  void insertItem(QGraphicsItem* item, int z = -1);
  /**
   * Returns ordered list of items.
   * @param order sort order: Qt::AscendingOrder or Qt::DescendingOrder.
   * @return ordered list.
   */
  QList<QGraphicsItem*> items(Qt::SortOrder order) const;
  /**
   * Maps from cm to pixel units.
   * @param p position in cm.
   * @return position in pixel units.
   */
  QPointF map(const QPointF& p) const;
  /**
   * Maps from pixel to cm units.
   * @param p position in pixel units.
   * @return position in cm units.
   */
  QPointF mapInv(const QPointF& p) const;
  /** Returns matrix for transformation from cm to pixel units. */
  inline const QTransform& matrix() const { return m_matrix; };
  /**
   * Prints plot.
   * @param printer pointer to printer object.
   */
  void print(QPrinter* printer);
  /**
   * Opens a data file. In case of success, a FrameItem and a corresponding
   * ArrayItem are constructed and added to the scene.
   * @param url URL of the file.
   * @param decimalChar decimal symbol of floating point numbers used in file.
   * @param autoPathHeader true for setting the heading to the path of the URL.
   * @param autoNorm true for automatic normalization.
   * @param logX true for logarithmic x axis.
   * @param logY true for logarithmic y axis.
   * @param x0 left page margin in cm.
   * @param y0 bottom page margin in cm.
   * @param width width in cm.0
   * @param height height in cm.
   * @param colorFrame color for frame and labels.
   * @param colorGrid color for grids and tics.
   * @param colorData color for data plot.
   * @param gridMode grid mode of frame item.
   * @param xText x axis text.
   * @param yText y axis text.
   * @param header header text.
   * @param relSize relative size of text, markers and lines.
   * @param colX x column index.
   * @param colY y column index.
   * @param colErr y error column index.
   * @param errorBars true to plot error bars.
   * @param symbol symbol type.
   * @return true when when succeeded.
   */
  bool readData(const QUrl& url,
                ArrayItem::DecimalCharacter decimalChar = ArrayItem::Point,
                bool autoPathHeader = false, bool autoNorm = true,
                bool logX = false, bool logY = false, double x0 = 4,
                double y0 = 3, double width = 15, double height = 10,
                unsigned colorFrame = 0, unsigned colorGrid = 0,
                unsigned colorData = 0,
                FrameItem::GridMode gridMode = FrameItem::AxesWithLabels,
                const QString& xText = "x", const QString& yText = "y",
                const QString& header = "", double relSize = 1, int colX = 0,
                int colY = 1, int colErr = 2, bool errorBars = false,
                KplItem::Symbol symbol = KplItem::SolidLine);
  /**
   * Reads plot file, initializes items and adds them to the scene.
   * @param url URL of the file.
   * @param window pointer to main window.
   * @return true for success.
   */
  bool readPlo(const QUrl& url, QWidget* window = 0);
  /**
   * Reads plot file and initializes plot items.
   * @param url URL of the file.
   * @param list list for plot items.
   * @param window pointer to main window.
   * @param colorFrame default color for frame and labels.
   * @param colorGrid default color for grids and tics.
   * @param colorData default color for data plot.
   * @param symbol default symbol type.
   * @param x0 default left page margin in cm.
   * @param y0 default bottom page margin in cm.
   * @param width default width in cm.
   * @param height default height in cm.
   * @param gridMode default frame grid mode.
   * @param colX default x column index.
   * @param colY default y column index.
   * @param colErr default y error column index.
   * @param errorBars default, true to plot error bars.
   * @return true for success.
   */
  static bool readPlo(const QUrl& url, QList<QGraphicsItem*>& list,
                      QWidget* window = 0, unsigned colorFrame = 0,
                      unsigned colorGrid = 0, unsigned colorData = 0,
                      KplItem::Symbol symbol = KplItem::OpenCircle,
                      double x0 = 4, double y0 = 3, double width = 15,
                      double height = 10,
                      FrameItem::GridMode gridMode = FrameItem::AxesWithLabels,
                      int colX = 0, int colY = 0, int colErr = 0,
                      bool errorBars = false);
  /**
   * Saves all plot items at URL.
   * @param url URL of file.
   * @param absolute save absolute URL when true.
   * @param format format character for floating point number presentation.
   * @param precision precision (number of digits) for floating point number
   * presentation.
   * @return true for success.
   */
  bool savePlo(const QUrl& url, bool absolute = true, char format = 'g',
               int precision = 6) const;
  /**
   * Saves plot items of a given list at URL.
   * @param url URL of file.
   * @param list list of items to save.
   * @param absolute save absolute URL when true.
   * @param format format character for floating point number presentation.
   * @param precision precision (number of digits) floating point number
   * presentation.
   * @return true for success.
   */
  static bool savePlo(const QUrl& url, const QList<QGraphicsItem*>& list,
                      bool absolute = true, char format = 'g',
                      int precision = 6);
  /**
   * Generates PostScript file.
   * @param url URL of the file (empty URL for preview of a temporary file).
   * @param landscape true for landscape orientation.
   * @param boundingBox true for calculation of bounding box.
   * @return true when file is saved.
   */
  bool savePS(const QUrl& url = QUrl(), bool landscape = false,
              bool boundingBox = true);
  /** Returns scene size in cm units. */
  inline const QSizeF size() const { return m_size; };
  /** Sets offsets for dashed lines. */
  void setDashOffsets() const;
  /**
   * Sets item properties according to transformation matrix.
   * @param item pointer to plot item.
   */
  void setProperties(QGraphicsItem* item) const;
  /**
   * Sets scene size and transformation matrix.
   * @param size size in cm units.
   */
  void setSize(const QSizeF& size);
  /**
   * Returns user unit steps for one pixel steps.
   * @param z z value of item, -1 for end of list.
   * @return user unit steps.
   */
  QPointF steps(int z = -1) const;
  /**
   * Emits signal #positionChanged.
   * @param item pointer to plot item.
   */
  void updatePos(QGraphicsItem* item);

signals:
  /**
   * Emitted when mouse has moved within the plot window.
   * @param pos mouse cursor position.
   * @param frame pointer to frame containing pos item or 0 if pos not within
   * a frame item.
   */
  void mouseMoved(const QPointF& pos, const FrameItem* frame);
  /**
   * Emitted when mouse button has been pressed.
   * @param pos mouse cursor position.
   * @param buttons combination of pressed mouse buttons.
   */
  void mousePressed(const QPointF& pos, Qt::MouseButtons buttons);
  /**
   * Emitted when position of item has changed.
   * @param item pointer to plot item.
   */
  void positionChanged(QGraphicsItem* item);
  /**
   * Emitted after creation of a preview file to allow scheduling for cleanup.
   * @param filename full path of preview file.
   */
  void previewCreated(const QString& filename);

protected:
  /**
   * Finds parent item corresponding to a z value of a plot item.
   * @param z z value of item, -1 for end of list.
   * @return pointer to parent frame or zero if no parent exists.
   */
  FrameItem* frame(int z = -1) const;
  /**
   * Emits signal #mouseMoved to indicate  actual position and the frame item
   * under the mouse cursor.
   * @param e mouse event.
   */
  virtual void mouseMoveEvent(QGraphicsSceneMouseEvent* e);
  /**
   * Emits signal #mousePressed to indicate  actual position and pressed button.
   * @param e mouse event.
   */
  virtual void mousePressEvent(QGraphicsSceneMouseEvent* e);
  double m_sx, m_sy;
  QTransform m_matrix, m_inv;
  QSizeF m_size;
};

#endif
