/***************************************************************************
                          legenditem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Tue Aug 31 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef LEGENDITEM_H
#define LEGENDITEM_H

#include <QGraphicsItem>
#include <QTextDocument>
#include "decoitem.h"

class KConfigGroup;

/**
  * Legend item class. Class for legends items explaining markers
  * and line types.
  * @author Werner Stille
  */
class Q_DECL_EXPORT LegendItem : public DecoItem, public QGraphicsItem
{
public:
  /** Specific QGraphicsItem type number for function type() */
  enum {Type = UserType + Legend
    /**< Specific QGraphicsItem type number of LegendItem for function
     * type().
     */
  };
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x0 x position in cm.
    * @param y0 y position in cm.
    * @param colorSymbol color for marker or line.
    * @param colorText color for text.
    * @param symbol symbol type.
    * @param xOff x offset for text in cm.
    * @param yOff y offset for text in cm.
    * @param parent pointer to parent item.
    * @param text legend text.
    * @param visible true for visible item.
    * @param relSize relative size of text.
    * @param symbolSize relative size of symbol.
    */
  LegendItem(double x0 = 0, double y0 = 0, unsigned colorSymbol = 0,
             unsigned colorText = 0, Symbol symbol = SolidLine,
             double xOff = 1, double yOff = 0, QGraphicsItem* parent = 0,
             const QString& text = QString(), bool visible = true,
             double relSize = 1, double symbolSize = 1);
  /**
    * Copy constructor.
    * @param a source item.
    */
  LegendItem(const LegendItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param colorSymbol default color for marker or line.
    * @param colorText default color for text.
    * @param symbol default symbol type.
    * @param frameWidth frame width in cm.
    * @param parent pointer to parent item.
    */
  LegendItem(const KConfigGroup* plo, unsigned colorSymbol, unsigned colorText,
             Symbol symbol, double frameWidth, QGraphicsItem* parent = 0);
  /** Destructor. */
  virtual ~LegendItem();
  /** Calculates bounding rectangle. */
  virtual QRectF boundingRect() const;
  /** Paints item. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /** Returns decoration pixmap for item list. */
  QPixmap pixmap() const;
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /** Sets size according to transformation matrix. */
  void setSize();
  /**
    * Sets symbol type.
    * @param symbol symbol type.
    */
  inline void setSymbol(Symbol symbol) { m_symbol = symbol; };
  /**
    * Sets symbol size.
    * @param symbolSize relative size.
    */
  inline void setSymbolSize(double symbolSize) { m_symbolSize = symbolSize; };
  /**
    * Sets text.
    * @param text text.
    */
  inline void setText(const QString& text) { m_text = text; };
  /**
    * Sets text color.
    * @param color color.
    */
  inline void setTextColor(unsigned color) { m_textColor = color; };
  /**
    * Sets x offset for text.
    * @param x x offset for text in cm.
    */
  inline void setTextX(double x) { m_textX = x; };
  /**
    * Sets y offset for text.
    * @param y y offset for text in cm.
    */
  inline void setTextY(double y) { m_textY = y; };
  /** Returns symbol type. */
  inline Symbol symbol() const { return m_symbol; };
  /** Returns relative symbol size. */
  inline double symbolSize() const { return m_symbolSize; };
  /** Returns text. */
  inline QString text() const { return m_text; };
  /** Returns text color. */
  inline unsigned textColor() const { return m_textColor; };
  /** Returns x offset for text in cm. */
  inline double textX() const { return m_textX; };
  /** Returns y offset for text in cm. */
  inline double textY() const { return m_textY; };
  /** Returns specific QGraphicsItem type number. */
  virtual int type() const { return Type; };
  /** Applies changes and forces redrawing. */
  void update();
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   */
  void writePlo(KConfigGroup* plo) const;

protected:
  /** Performs initialization. */
  void init();
  /** Updates position. */
  virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent* event);
  /** Sets properties according to transformation matrix. */
  void setProperties();
  /** Returns text rectangle. */
  QRectF textRect() const;
  Symbol m_symbol;
  unsigned m_textColor;
  double m_textX, m_textY, m_symbolSize;
  QString m_text;
  QTextDocument m_doc;
  QPen m_pen;
  QRectF symbRect;
};

#endif
