/***************************************************************************
                          scalebaritem.cpp  -  description
                             -------------------
    begin                : Don Jun 12 2003
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QAbstractTextDocumentLayout>
#include <QPainter>
#include <QTextFrame>
#include <KConfigGroup>
#include "kplscene.h"
#include "scalebaritem.h"
#include "textitem.h"

ScaleBarItem::ScaleBarItem(double x0, double y0, unsigned color,
                           QGraphicsItem* parent, double length,
                           const QString& text, bool visible, bool vertical,
                           double relSize) :
 DecoItem(color, x0, y0, relSize), QGraphicsItem(parent),
 m_vertical(vertical), m_length(length), m_text(text)
{
  if (!visible)
    hide();
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  init();
}

ScaleBarItem::ScaleBarItem(const ScaleBarItem& a) :
 DecoItem(a), QGraphicsItem(), m_vertical(a.m_vertical), m_length(a.m_length),
 l2(a.l2), wRel(a.wRel), m_text(a.m_text)
{
  setZValue(a.zValue());
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  setSelected(a.isSelected());
  setVisible(a.isVisible());
}

ScaleBarItem::ScaleBarItem(const KConfigGroup* plo, unsigned color,
                           QGraphicsItem* parent) :
 DecoItem(plo, color), QGraphicsItem(parent)
{
  if (!plo->readEntry("active", true))
    hide();
  m_vertical = plo->readEntry("or", 0);
  m_length = plo->readEntry("len", 1.0);
  m_text = plo->readEntry("s", "");
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  init();
}

ScaleBarItem::~ScaleBarItem()
{
}

QRectF ScaleBarItem::boundingRect() const
{
  double pw = m_pen.widthF();
  double w = l2 + pw;
  double h = 0.01 * wRel + pw;
  double d = -0.5 * pw;
  return textRect().united(m_vertical ?
                           QRectF(-0.5 * h, -d, h, -w) :
                           QRectF(d, -0.5 * h, w, h));
}

void ScaleBarItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*,
                         QWidget*)
{
  if (isSelected()) {
    painter->setPen(QPen(Qt::DashLine));
    painter->drawRect(boundingRect());
  }
  painter->setPen(m_pen);
  double dy = 0.005 * wRel;
  if (m_vertical)
    painter->rotate(-90);
  painter->drawLine(QPointF(), QPointF(l2, 0));
  painter->drawLine(QPointF(0, -dy), QPointF(0, dy));
  painter->drawLine(QPointF(l2, -dy), QPointF(l2, dy));
  if (m_vertical)
    painter->rotate(90);
  painter->translate(textRect().topLeft());
  m_doc.drawContents(painter);
}

QPixmap ScaleBarItem::pixmap() const
{
  return KplItem::pixmap(m_pen);
}

void ScaleBarItem::setProperties(const QTransform& m)
{
  m_matrix = m;
  setProperties();
}

void ScaleBarItem::setSize()
{
  m_pen.setWidthF(penWidth(this, FramePen));
  wRel = parentWidth(this) * m_relSize;
  QFont f("Helvetica");
  f.setPixelSize(qRound(0.033333 * wRel));
  m_doc.setDefaultFont(f);
}

void ScaleBarItem::writePlo(KConfigGroup* plo) const
{
  plo->writeEntry("Type", "SCALEBARITEM");
  DecoItem::writePlo(plo, isVisible());
  plo->writeEntry("or", m_vertical ? 1 : 0);
  plo->writeEntry("len", m_length);
  plo->writeEntry("s", m_text);
}

void ScaleBarItem::update()
{
  init();
  prepareGeometryChange();
  setProperties();
  QGraphicsItem::update();
}

void ScaleBarItem::init()
{
  TextItem::setText(m_doc, m_text, m_color);
  QTextFrameFormat tf = m_doc.rootFrame()->frameFormat();
  tf.setMargin(0);
  m_doc.rootFrame()->setFrameFormat(tf);
}

void ScaleBarItem::mouseReleaseEvent(QGraphicsSceneMouseEvent* event)
{
  QGraphicsItem::mouseReleaseEvent(event);
  ((KplScene*)scene())->updatePos(this);
}

void ScaleBarItem::setProperties()
{
  setPos(m_matrix.map(QPointF(m_x0, m_y0)));
  m_pen = createPen(this, m_color);
  setSize();
  l2 = m_matrix.m11() * m_length;
}

QRectF ScaleBarItem::textRect() const
{
  QRectF r = m_doc.documentLayout()->blockBoundingRect(m_doc.begin());
  if (m_vertical)
    r.moveTopLeft(r.topLeft() +
                  QPointF(0.01 * wRel, -0.5 * l2 -
                          0.65 * QFontMetricsF(m_doc.defaultFont()).ascent()));
  else
    r.moveTopLeft(r.topLeft() +
                  QPointF(0.5 * (l2 - r.width()),
                          -QFontMetricsF(m_doc.defaultFont()).ascent()
                          - 0.015 * wRel));
  return r;
}
