//
// C++ Interface: scaleditem
//
// Description: 
//
//
// Author: Werner Stille <kpl@werner-stille.de>, (C) 2017
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef SCALEDITEM_H
#define SCALEDITEM_H

#include <QGraphicsItem>
#include <QPainterPath>
#include <QRectF>
#include "kplitem.h"

/**
  * Scaled item class. Base class for array and function items.
  * @author Werner Stille
  */
class Q_DECL_EXPORT ScaledItem : public KplItem, public QGraphicsItem
{
public:
  /** Constructor. */
  ScaledItem();
  /**
    * Copy constructor.
    * @param a source item.
    */
  ScaledItem(const ScaledItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param symbol default symbol type.
    * @param color default color.
    * @param parent pointer to parent item.
    */
  ScaledItem(const KConfigGroup* plo, Symbol symbol, unsigned color,
             QGraphicsItem* parent);
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param fillStyle fill style.
    * @param symbol symbol type.
    * @param color color.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param xOffset offset for x values.
    * @param yOffset offset for x values.
    * @param relSize relative size.
    * @param parent pointer to parent item.
    */
  ScaledItem(FillStyle fillStyle, Symbol symbol, unsigned color, double fx,
             double fy, double xOffset, double yOffset, double relSize,
             QGraphicsItem* parent);
  /** Destructor. */
  virtual ~ScaledItem();
  /** Returns color. */
  inline unsigned color() const { return m_color; };
  /**
    * Calculates normalization factor.
    * @param a absolute double value.
    * @param iExp reference to logarithm of normalization factor.
    * @param f reference to normalization factor.
    */
  static void exponent(double a, int& iExp, double& f);
  /** Returns fill style. */
  inline FillStyle fillStyle() const { return m_fillStyle; };
  /**
    * Checks if both normalization factors are equal to given values.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @return true when both normalization factors are equal to given values.
    */
  inline bool isNormalized(double fx, double fy) const
    { return (m_normX == fx) && (m_normY == fy); };
  /**
    * Calculates minimum and maximum values of an array.
    * @param xMin reference to minimum value.
    * @param xMax reference to maximum value.
    * @param values double array.
    * @param n size of array.
    * @return true for xmax > xmin.
    */
  static bool limits(double& xMin, double& xMax, const double* values, int n);
  /**
    * Sets normalization factors.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    */
  inline void normalize(double fx, double fy)  { m_normX = fx; m_normY = fy; };
  /** Returns normalization factor for x values. */
  inline double normX() const { return m_normX; };
  /** Returns normalization factor for y values. */
  inline double normY() const { return m_normY; };
  /** Returns decoration pixmap for item list. */
  QPixmap pixmap() const;
  /**
    * Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param symbol default symbol type.
    * @param color default color.
    */
  void readPlo(const KConfigGroup* plo, Symbol symbol, unsigned color);
  /**
    * Sets color.
    * @param color color.
    */
  inline void setColor(unsigned color) { m_color = color; };
  /**
    * Sets fill style.
    * @param style fill style.
    */
  inline void setFillStyle(FillStyle style) { m_fillStyle = style; };
  /**
    * Sets normalization factor for x values.
    * @param fx normalization factor.
    */
  inline void setNormX(double fx) { m_normX = fx; };
  /**
    * Sets normalization factor for y values.
    * @param fy normalization factor.
    */
  inline void setNormY(double fy) { m_normY = fy; };
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /** Sets size according to transformation matrix. */
  void setSize();
  /**
    * Sets symbol type.
    * @param symbol symbol type.
    */
  inline void setSymbol(Symbol symbol) { m_symbol = symbol; };
  /**
    * Sets offset for x values.
    * @param xOffset offset.
    */
  inline void setXOffset(double xOffset) { m_xOffset = xOffset; };
  /**
    * Sets offset for y values.
    * @param yOffset offset.
    */
  inline void setYOffset(double yOffset) { m_yOffset = yOffset; };
  /** Returns symbol type. */
  inline Symbol symbol() const { return m_symbol; };
  /** Applies changes and forces redrawing. */
  void update();
  /** Updates geometry. */
  void updateGeometry();
  /**
    * Writes into plot file.
    * @param plo KConfigGroup object.
    * @param visible true for visible item.
    */
  void writePlo(KConfigGroup* plo, bool visible) const;
  /** Returns offset for x values. */
  inline double xOffset() const { return m_xOffset; };
  /** Returns offset for y values. */
  inline double yOffset() const { return m_yOffset; };

protected:
  /**
    * Calculates line length.
    * @param x array with x values.
    * @param y array with y values.
    * @param n number of values.
    * @return length in pixel units.
    */
  double length(const double* x, const double* y, int n) const;
  /**
    * Plots array.
    * @param painter pointer to QPainter object, 0 for calculating bounding
    * box.
    * @param x array with x values.
    * @param y array with y values.
    * @param n number of values.
    */
  void plotArray(QPainter* painter, const double* x, const double* y,
                 int n) const;
  /** Sets properties according to transformation matrix. */
  void setProperties();
  FillStyle m_fillStyle;
  Symbol m_symbol;
  unsigned m_color;
  double m_normX, m_normY, m_xOffset, m_yOffset, dydx, myDashOffset;
  mutable QPen m_pen;
  mutable QPainterPath bPath;
  mutable QRectF bRect;
  QRectF symbRect;
  QTransform m_matrix;
};

#endif
