/***************************************************************************
                          threeditem.h  -  description
                             -------------------
    begin                : Tue Dec 25 2001
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef THREEDITEM_H
#define THREEDITEM_H

#include <QTransform>
#include "frameitem.h"

class PSGraph;

/**
  * 3D item class. Base class for 3D array items and 3d function items.
  * @author Werner Stille
  */
class Q_DECL_EXPORT ThreeDItem : public FrameItem
{
public:
  /** 3D mode. Determines how the surface y = f(x, z) should be plotted. */
  enum Mode3D {
    AlongX, ///< Lines along x direction.
    AlongZ, ///< Lines along z direction.
    Grid    ///< x, z grid.
  };
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x0 left page margin in centimeters.
    * @param y0 bottom page margin in centimeters.
    * @param width width in centimeters.
    * @param height height in centimeters.
    * @param colFrame color for frame and labels.
    * @param colGrid color for grids and tics.
    * @param colData color for data plot.
    * @param logX true for logarithmic x axis.
    * @param logY true for logarithmic y axis.
    * @param logZ true for logarithmic y axis.
    * @param gridMode grid mode.
    * @param mode3D 3D mode.
    * @param xMax maximum x value.
    * @param yMax maximum y value.
    * @param zMax maximum z value.
    * @param textX x axis text.
    * @param textY y axis text.
    * @param textZ z axis text.
    * @param header heading text.
    * @param relSize relative size of text, markers and lines.
    * @param numDigX number of digits for x axis.
    * @param numDigY number of digits for y axis.
    * @param numDigZ number of digits for z axis.
    * @param minTicX number of intervals between major tics of x axis.
    * @param minTicY number of intervals between major tics of y axis.
    * @param minTicZ number of intervals between major tics of z axis.
    * @param expX logarithm of normalization factor for x axis.
    * @param expY logarithm of normalization factor for y axis.
    * @param expZ logarithm of normalization factor for z axis.
    * @param xMin minimum x value.
    * @param yMin minimum y value.
    * @param zMin minimum z value.
    * @param xTic distance between major tics of x axis.
    * @param yTic distance between major tics of y axis.
    * @param zTic distance between major tics of z axis.
    * @param deltaX x step size of grid.
    * @param deltaZ z step size of grid.
    * @param phi angle of rotation about y axis in deg.
    * @param theta tilt angle of y axis in deg.
    * @param boxX x box width in cm.
    * @param boxY y box width in cm.
    * @param boxZ z box width in cm.
    * @param xRef x position of center of box bottom plane.
    * @param yRef y position of center of box bottom plane.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param fz normalization factor for z values.
    * @param showFrame true to show frame.
    * @param collapsed true for collapsed in list view.
    * @param visible true for visible item.
    */
  ThreeDItem(double x0 = 4, double y0 = 3, double width = 15,
             double height = 10, unsigned colFrame = 0, unsigned colGrid = 0,
             unsigned colData = 0, bool logX = false, bool logY = false,
             bool logZ = false, GridMode gridMode = AxesWithLabels,
             Mode3D mode3D = Grid, double xMax = 15, double yMax = 10,
             double zMax = 15, const QString& textX = "",
             const QString& textY = "", const QString& textZ = "",
             const QString& header = "", double relSize = 1, int numDigX = -1,
             int numDigY = -1, int numDigZ = -1, int minTicX = 5,
             int minTicY = 2, int minTicZ = 5, int expX = 0, int expY = 0,
             int expZ = 0, double xMin = 0, double yMin = 0, double zMin = 0,
             double xTic = 5, double yTic = 2, double zTic = 5,
             double deltaX = 0, double deltaZ = 0, double phi = 20,
             double theta = -20, double boxX = 7.5, double boxY = 7.5,
             double boxZ = 7.5, double xRef = 7.5, double yRef = 4,
             double fx = 1, double fy = 1, double fz = 1,
             bool showFrame = true, bool collapsed = false,
             bool visible = true);
  /**
    * Copy constructor.
    * @param a source item.
    */
  ThreeDItem(const ThreeDItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param x0 default value for left page margin in centimeters.
    * @param y0 default value for bottom page margin in centimeters.
    * @param width default value for width in centimeters.
    * @param height default value for height in centimeters.
    * @param colorFrame default value for color for frame and labels.
    * @param colorGrid default value for color for grids and tics.
    * @param colorData color for data plot.
    * @param gridMode grid mode.
    */
  ThreeDItem(const KConfigGroup* plo, double x0, double y0, double width,
             double height, unsigned colorFrame, unsigned colorGrid,
             unsigned colorData, GridMode gridMode);
  /** Destructor. */
  ~ThreeDItem();
  /** Calculates bounding rectangle. */
  virtual QRectF boundingRect() const;
  /** Returns x box width in cm. */
  inline double boxX() const { return m_boxX; };
  /** Returns y box width in cm. */
  inline double boxY() const { return m_boxY; };
  /** Returns z box width in cm. */
  inline double boxZ() const { return m_boxZ; };
  /** Returns color for data. */
  inline unsigned colorData() const { return m_colorData; };
  /** Returns x step size of grid. */
  inline double deltaX() const { return m_deltaX; };
  /** Returns z step size of grid. */
  inline double deltaZ() const { return m_deltaZ; };
  /**
   * Exports value table to file.
   * @param stream stream for output file.
   * @param separator column separator character.
   * @param format C format character.
   * @param precision precision.
   */
  void exportTable(QTextStream& stream, char separator, char format,
                   int precision) const;
  /** Returns logarithm of normalization factor for z axis. */
  inline int expZ() const { return m_expZ; };
  /** Returns normalization factor for x values. */
  inline double fX() const { return m_fX; };
  /** Returns normalization factor for y values. */
  inline double fY() const { return m_fY; };
  /** Returns normalization factor for z values. */
  inline double fZ() const { return m_fZ; };
  /** Returns true for logarithmic z axis. */
  inline bool logZ() const { return m_logZ; };
  /** Returns number of intervals between major tics of z axis. */
  inline int minTicZ() const { return m_minTicZ; };
  /**
    * Returns 0 for lines parallel to x, 1 for lines parallel to z,
    * 2 for x,z grid.
    */
  inline Mode3D mode3D() const { return m_mode3D; };
  /** Returns number of digits for z axis. */
  inline int numDigZ() const { return m_numDigZ; };
  /** Paints item. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /**
    * Paints or calculates bounding area.
    * @param painter QPainter object, 0 for calculating bounding area.
    * @param ps pointer to PSGraph object, != 0 for PostScript output.
    */
  void paintCond(QPainter* painter, PSGraph* ps = 0) const;
  /** Returns angle of rotation about y axis in deg. */
  inline double phi() const { return m_phi; };
  /**
    * Sets x box width in cm.
    * @param width width.
    */
  inline void setBoxX(double width) { m_boxX = width; };
  /**
    * Sets y box width in cm.
    * @param width width.
    */
  inline void setBoxY(double width) { m_boxY = width; };
  /**
    * Sets z box width in cm.
    * @param width width.
    */
  inline void setBoxZ(double width) { m_boxZ = width; };
  /**
    * Sets color for data.
    * @param color color.
    */
  inline void setColorData(unsigned color) { m_colorData = color; };
  /**
    * Sets x step size of grid.
    * @param dx step size.
    */
  inline void setDeltaX(double dx) { m_deltaX = dx; };
  /**
    * Sets z step size of grid.
    * @param dz step size.
    */
  inline void setDeltaZ(double dz) { m_deltaZ = dz; };
  /**
    * Sets exponent of z normalization factor.
    * @param e exponent.
    */
  inline void setExpZ(int e) { m_expZ = e; };
  /**
    * Sets x normalization factor.
    * @param fx factor.
    */
  inline void setFX(double fx) { m_fX = fx; };
  /**
    * Sets y normalization factor.
    * @param fy factor.
    */
  inline void setFY(double fy) { m_fY = fy; };
  /**
    * Sets z normalization factor.
    * @param fz factor.
    */
  inline void setFZ(double fz) { m_fZ = fz; };
  /**
    * Sets logarithmic mode of z axis
    * @param logZ true for logarithmic z axis.
    */
  inline void setLogZ(bool logZ) { m_logZ = logZ; };
  /**
    * Sets number of intervals between major tics of z axis.
    * @param minTicZ number of intervals.
    */
  inline void setMinTicZ(int minTicZ) { m_minTicZ = minTicZ; };
  /**
    * Sets 3D mode.
    * @param mode3D 3D mode.
    */
  inline void setMode3D(Mode3D mode3D) { m_mode3D = mode3D; };
  /**
    * Sets number of digits for z axis.
    * @param numDigZ number of digits.
    */
  inline void setNumDigZ(int numDigZ) { m_numDigZ = numDigZ; };
  /**
    * Sets angle of rotation about y axis.
    * @param phi angle in deg.
    */
  inline void setPhi(double phi) { m_phi = phi; };
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /**
    * Sets frame mode.
    * @param showFrame true to show frame.
    */
  inline void setShowFrame(bool showFrame) { m_showFrame = showFrame; };
  /**
    * Sets z axis text.
    * @param text axis text.
    */
  inline void setTextZ(const QString& text) { m_textZ = text; };
  /**
    * Sets angle of rotation about y axis.
    * @param theta angle in deg.
    */
  inline void setTheta(double theta) { m_theta = theta; };
  /**
    * Sets x position of center of box bottom plane.
    * @param xRef x position in cm.
    */
  inline void setXRef(double xRef) { m_xRef = xRef; };
  /**
    * Sets y position of center of box bottom plane.
    * @param yRef y position in cm.
    */
  inline void setYRef(double yRef) { m_yRef = yRef; };
  /**
    * Sets upper z data limit.
    * @param zMax upper limit in user units.
    */
  inline void setZMax(double zMax) { m_zMax = zMax; };
  /**
    * Sets lower z data limit.
    * @param zMin lower limit in user units.
    */
  inline void setZMin(double zMin) { m_zMin = zMin; };
  /**
    * Sets z tic distance.
    * @param zTic distance between major tics of z axis.
    */
  inline void setZTic(double zTic) { m_zTic = zTic; };
  /** Returns true for shown frame. */
  inline bool showFrame() const { return m_showFrame; };
  /** Returns z axis text. */
  inline QString textZ() const { return m_textZ; };
  /** Returns angle of rotation about y axis in deg. */
  inline double theta() const { return m_theta; };
  /** Applies changes and forces redrawing. */
  void update();
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   */
  void writePlo(KConfigGroup* plo) const;
  /** Returns x position of center of box bottom plane. */
  inline double xRef() const { return m_xRef; };
  /** Returns y position of center of box bottom plane. */
  inline double yRef() const { return m_yRef; };
  /** Returns pointer to y value array. */
  double** yValues() const { return m_yValues; };
  /** Returns upper z data limit user units. */
  inline double zMax() const { return m_zMax; };
  /** Returns lower z data limit user units. */
  inline double zMin() const { return m_zMin; };
  /** Returns distance between major tics of z axis. */
  inline double zTic() const { return m_zTic; };

protected:
  /** 3D mode. Determines how the surface y = f(x, z) should be plotted. */
  enum Axis {
    X, ///< x axis.
    Y, ///< y axis.
    Z  ///< z axis.
  };
  /**
    * Plots 3D axis.
    * @param painter pointer to QPainter object.
    * @param vMin minimum axis value.
    * @param vMax maximum axis value.
    * @param tic distance between major tics.
    * @param minTic number of intervals between major tics.
    * @param f normalization factor.
    * @param scale scaling factor.
    * @param xAxis 3D x position of axis start point.
    * @param zAxis 3D z position of axis start point.
    * @param ticLen tic length.
    * @param numDig number of digits.
    * @param axis axis.
    * @param log true for logarithmic axis.
    * @param ps pointer to PSGraph object, != 0 for PostScript output.
    */
  void axis3D(QPainter* painter, double vMin, double vMax, double tic,
              int minTic,
              double f, double scale, double xAxis, double zAxis,
              double ticLen, int numDig, Axis axis, bool log,
              PSGraph* ps = 0) const;
  /**
    * Plots text.
    * @param painter pointer to QPainter object.
    * @param text text.
    * @param x x position in pixel units.
    * @param y y position in pixel units.
    * @param alignment horizontal alignment.
    * @param ps pointer to PSGraph object, != 0 for PostScript output.
    * @param dir text direction in degree.
    */
  void drawText(QPainter* painter, const QString& text, double x, double y,
                TextItem::Alignment alignment, PSGraph* ps,
                double dir = 0) const;
  /**
    * Plots 3D frame.
    * @param painter pointer to QPainter object.
    * @param iBack index of vertex furthest in the background.
    * @param ps pointer to PSGraph object, != 0 for PostScript output.
    */
  void frame3D(QPainter* painter, int iBack, PSGraph* ps = 0) const;
  /**
    * Draws line in 3D space.
    * @param painter pointer to QPainter object.
    * @param x1 3D x value of point 1.
    * @param y1 3D y value of point 1.
    * @param z1 3D z value of point 1.
    * @param x2 3D x value of point 2.
    * @param y2 3D y value of point 2.
    * @param z2 3D z value of point 2.
    * @param ps pointer to PSGraph object, != 0 for PostScript output.
    */
  void line3D(QPainter* painter, double x1, double y1, double z1,
              double x2, double y2, double z2, PSGraph* ps = 0) const;
  /**
    * Plots 3D data.
    * @param painter pointer to QPainter object.
    * @param ps pointer to PSGraph object, != 0 for PostScript output.
    */
  void plotArray3D(QPainter* painter, PSGraph* ps = 0) const;
  /**
    * Plots curve in 3D space.
    * @param painter pointer to QPainter object.
    * @param xArray true for array in xData.
    * @param yArray true for array in yData.
    * @param zArray true for array in zData.
    * @param xData array or start value of x data.
    * @param yData array or start value of y data.
    * @param zData array or start value of z data.
    * @param xScale factor for x array step width x data.
    * @param yScale factor for y array step width y data.
    * @param zScale factor for z array step width z data.
    * @param firstLine true for first line of 3D surface.
    * @param n number of points of the curve.
    * @param xOffset 2D x offset for 3D y axis mimimum.
    * @param yOffset 2D y offset for 3D y axis mimimum.
    * @param vertex array to store vertices.
    * @param ps pointer to PSGraph object, != 0 for PostScript output.
    */
  void plot3D(QPainter* painter, bool xArray, bool yArray, bool zArray,
              const double* xData, const double* yData,
              const double* zData, double xScale, double yScale,
              double zScale, bool firstLine, int n, double xOffset,
              double yOffset, double* vertex, PSGraph* ps = 0) const;
  /**
    * Plots array.
    * @param painter pointer to QPainter object, 0 for calculating bounding
    * box.
    * @param x array with x values.
    * @param y array with y values.
    * @param n number of values.
    * @param ps pointer to PSGraph object, != 0 for PostScript output.
    */
  void plotArray(QPainter* painter, const double* x, const double* y, int n,
                 PSGraph* ps = 0) const;
  /**
    * Draws frame and axes of 3D plot.
    * @param painter pointer to QPainter object.
    * @param ps pointer to PSGraph object, != 0 for PostScript output.
    */
  void raster3d(QPainter* painter, PSGraph* ps = 0) const;
  /** Sets properties according to transformation matrix. */
  void setProperties();
  /**
    * Calculates 2D x value.
    * @param x 3D x value.
    * @param z 3D z value.
    */
  double xFrom3d(double x, double z) const;
  /**
    * Calculates 2D y value.
    * @param x 3D x value.
    * @param y 3D y value.
    * @param z 3D z value.
    */
  double yFrom3d(double x, double y, double z) const;
  bool m_showFrame, m_logZ;
  mutable bool logXo, logZo;
  int m_numDigZ, m_minTicZ, m_expZ;
  Mode3D m_mode3D;
  mutable int nx, nz;
  unsigned m_colorData;
  double m_zMin, m_zMax, m_deltaX, m_deltaZ, m_zTic, m_phi, m_theta, m_boxX,
         m_boxY, m_boxZ, m_xRef, m_yRef, m_fX, m_fY, m_fZ, zminl;
  mutable double xStep, zStep, xMino, xMaxo, deltaXo, zMino, zMaxo, deltaZo,
                 scalx, scaly, scalz, scal3D, phio, thetao, xOff3D, yOff3D;
  mutable double** m_yValues;
  mutable double vertex[16];
  mutable QTransform m3D;
  mutable QVector<double> yMin3D, yMax3D;
  QString m_textZ;
};

#endif
