/***************************************************************************
                          arraydlg.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Apr 25 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ARRAYDLG_H
#define ARRAYDLG_H

#include <QString>
#include <QDialog>
#include "arrayitem.h"

class QCheckBox;
class QLabel;
class QPushButton;
class QSpinBox;
class QToolButton;
class QUrl;
class KColorButton;
class ArrayItem;
class BrushButton;
class KplDoc;
class DoubleEdit;
class DoubleSpinBox;
class SpinBox;
class SymbolButton;

/**
  * Array settings dialog class. Provides a dialog for the settings
  * of an array plot item.
  * @author Werner Stille
  */
class ArrayDlg : public QDialog
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param parent pointer to parent widget.
   * @param model pointer to KplDoc model object.
   * @param ai pointer to array object.
   * @param pos dialog position.
   */
  ArrayDlg(QWidget* parent, KplDoc* model, ArrayItem* ai, const QPoint& pos);
  /** Destructor. */
  ~ArrayDlg();

protected:
  /**
   * Applies settings.
   * @param ok true to close the dialog.
   */
  void apply(bool ok = false);
  /**
   * Reimplemented to display help when key F1 is pressed.
   * @param e Key event.
   */
  virtual void keyPressEvent(QKeyEvent* e);
  /**
   * Loads file.
   * @param url URL of file.
   * @param idec decimal character: Point or Comma.
   * @return true when when succeeded.
   */
  bool load(const QUrl& url, ArrayItem::DecimalCharacter idec);
  /**
   * Reads edited data into array object.
   * @param fromBuffer read editor buffer instead of file when true.
   * @return true when when succeeded.
   */
  bool readData(bool fromBuffer = false);
  /**
   * Updates ranges of spin boxes.
   * @param ad pointer to array object.
   */
  void setSpinBoxes(const ArrayItem* ad);
  QString buf;
  KplDoc* m;
  ArrayItem *ai, *ait;
  DoubleEdit *efx, *efy;
  DoubleSpinBox *ex0, *ey0;
  QLabel* fileName;
  QDialogButtonBox* buttonBox;
  QSpinBox *ix, *iy, *ie, *iStart, *n;
  SpinBox *symb, *eRelSize;
  QCheckBox *err, *internal;
  BrushButton *fill;
  SymbolButton* bSymb;
  KColorButton* colData;
  QPushButton *edit;
  QToolButton *bReload;

protected slots:
  /** Applies settings without leaving the dialog. */
  void slotApply();
  /**
   * Handles button clicks.
   * @param button pointer to button.
   */
  virtual void slotButtonClicked(QAbstractButton* button);
  /** Edits internal data. */
  void slotEdit();
  /** Enables buttons. */
  void slotEdited();
  /** Applies edited array data. */
  void slotEditApply();
  /** Displays the dialog for selection of a data file. */
  void slotFile();
  /**
   * Adjusts state of widgets when fill pattern is changed.
   * @param brush new pattern number.
   */
  void slotFillChanged(int brush);
  /** Toggles internal/external data. */
  void slotInternal(bool state);
  /**
   * Adjusts the Start index spinbox when number of points is changed.
   * @param n new number of points.
   */
  void slotNChanged(int n);
  /** Reloads file. */
  void slotReload();
  /** Restores window size. */
  void slotRestoreSize();
  /**
   * Adjusts the Number of points spinbox when start index is changed.
   * @param i new start index.
   */
  void slotStartChanged(int i);
};

#endif
