/***************************************************************************
                          arrowdlg.cpp  -  description
                             -------------------
    begin                : Sat Apr 14 2001
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <math.h>
#include <QtMath>
#include <QButtonGroup>
#include <QCheckBox>
#include <QDesktopServices>
#include <QDialogButtonBox>
#include <QDoubleSpinBox>
#include <QLabel>
#include <QLayout>
#include <QRadioButton>
#include <QUndoStack>
#include <KColorButton>
#include <KLocalizedString>
#include "arrowdlg.h"
#include "arrowitem.h"
#include "kpldoc.h"
#include "spinbox.h"
#include "symboldlg.h"

ArrowDlg::ArrowDlg(QWidget* parent, KplDoc* model, ArrowItem* ai0,
                   const QPoint& pos) :
 DecoDlg(parent, i18n("Arrow"), model, ai0, pos), koord(0), ai(ai0)
{
  initPolarEntries(ai, grid, &eDir, &eLen);
  grid->addWidget(new QLabel(i18n("Coordinates")), 1, 0);
  QHBoxLayout* hbox = new QHBoxLayout;
  grid->addLayout(hbox, 1, 1, 1, 6);
  hbox->addItem(new QSpacerItem(10, 10));
  QButtonGroup* bg = new QButtonGroup;
  QRadioButton* rb;
  int i = 0;
  hbox->addWidget(rb = new QRadioButton(i18n("Head")));
  bg->addButton(rb, i++);
  rb->setChecked(true);
  hbox->addWidget(rb = new QRadioButton(i18n("Center")));
  bg->addButton(rb, i++);
  hbox->addWidget(rb = new QRadioButton(i18n("End")));
  bg->addButton(rb, i);
  hbox->addItem(new QSpacerItem(20, 10, QSizePolicy::Expanding));
  initEntries(2);
  grid->addWidget(new QLabel(i18n("Type")), 4, 0);
  grid->addWidget(symb = new SpinBox(ai->symbol(), 1, 9), 4, 1);
  grid->addWidget(new SymbolButton(symb, true, false), 4, 2);
  grid->addWidget(dbl = new QCheckBox(i18n("Double headed")), 4, 4, 1, 1);
  dbl->setChecked(ai->doubleHeaded());
  connect(bg, SIGNAL(buttonClicked(int)), SLOT(slotSetCoord(int)));
  connect(eDir, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  connect(eLen, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  connect(symb, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(dbl, SIGNAL(toggled(bool)), SLOT(slotEdited()));
  connect(buttonBox, SIGNAL(clicked(QAbstractButton*)),
          SLOT(slotButtonClicked(QAbstractButton*)));
}

ArrowDlg::~ArrowDlg()
{
}

void ArrowDlg::initPolarEntries(const VectorItem* ai, QGridLayout* gr,
                                QDoubleSpinBox** eD, QDoubleSpinBox** eL)
{
  gr->addWidget(new QLabel(i18n("Direction")), 0, 0);
  gr->addWidget(*eD = new QDoubleSpinBox, 0, 1);
  (*eD)->setRange(-360, 360);
  (*eD)->setSuffix(QString::fromUtf8("\302\260"));
  (*eD)->setValue(ai->dir());
  gr->addWidget(new QLabel(i18n("Length")), 0, 3);
  gr->addWidget(*eL = new QDoubleSpinBox, 0, 4);
  (*eL)->setMaximum(200);
  (*eL)->setSingleStep(0.1);
  (*eL)->setSuffix(" cm");
  (*eL)->setValue(ai->length());
}

void ArrowDlg::calcCoord(double* x, double* y, double dx, double dy)
{
  switch (koord) {
    case 1:
      *x += 0.5 * dx;
      *y += 0.5 * dy;
      break;
    case 2:
      *x += dx;
      *y += dy;
  }
}

void ArrowDlg::slotButtonClicked(QAbstractButton* button)
{
  int id = buttonBox->standardButton(button);
  switch (id) {
    case QDialogButtonBox::Apply:
      buttonBox->button(QDialogButtonBox::Apply)->setEnabled(false);
      buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
    case QDialogButtonBox::Ok: {
      ArrowItem* aiNew = new ArrowItem(*ai);
      getValues(aiNew);
      aiNew->setSymbol(KplItem::Symbol(symb->value()));
      aiNew->setDoubleHeaded(dbl->isChecked());
      double dir = eDir->value();
      aiNew->setDir(dir);
      double len = eLen->value();
      aiNew->setLength(len);
      dir = qDegreesToRadians(dir);
      double x0 = aiNew->x0();
      double y0 = aiNew->y0();
      calcCoord(&x0, &y0, len * cos(dir), len * sin(dir));
      aiNew->setX0(0.01 * qRound(100 * x0));
      aiNew->setY0(0.01 * qRound(100 * y0));
      m->undoStack()->push(new EditCommand(aiNew, ai, m));
      ai = aiNew;
      if (id == QDialogButtonBox::Ok)
        accept();
      break;
    }
    case QDialogButtonBox::Help:
      QDesktopServices::openUrl(QUrl(QStringLiteral("help:/kpl/arrow.html")));
      break;
    case QDialogButtonBox::Cancel:
      reject();
  }
}

void ArrowDlg::slotSetCoord(int id)
{
  if (id != koord) {
    double dir = qDegreesToRadians(eDir->value());
    double len = eLen->value();
    double dx = len * cos(dir);
    double dy = len * sin(dir);
    double x0 = ex0->value();
    double y0 = ey0->value();
    calcCoord(&x0, &y0, dx, dy);
    koord = id;
    calcCoord(&x0, &y0, -dx, -dy);
    disconnect(ex0, SIGNAL(valueChanged(const QString&)),
               this, SLOT(slotEdited()));
    ex0->setValue(x0);
    connect(ex0, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
    disconnect(ey0, SIGNAL(valueChanged(const QString&)), this,
               SLOT(slotEdited()));
    ey0->setValue(y0);
    connect(ey0, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  }
}
