/***************************************************************************
                          edittextdlg.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Apr 25 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QButtonGroup>
#include <QDesktopServices>
#include <QDialogButtonBox>
#include <QGroupBox>
#include <QKeyEvent>
#include <QLayout>
#include <QLineEdit>
#include <QPainter>
#include <QPushButton>
#include <QRadioButton>
#include <QTextFrame>
#include <QTimer>
#include <QToolButton>
#include <KConfigGroup>
#include <KLocalizedString>
#include <KSharedConfig>
#include <KWindowConfig>
#include "edittextdlg.h"
#include "textitem.h"

Preview::Preview(QWidget* parent) : QWidget(parent)
{
  setAutoFillBackground(true);
  setBackgroundRole(QPalette::Base);
  QFont f("Helvetica");
  f.setPixelSize(20);
  m_doc.setDefaultFont(f);
}

Preview::~Preview()
{
}

void Preview::setText(const QString& s)
{
  TextItem::setText(m_doc, s);
  QTextFrameFormat tf = m_doc.rootFrame()->frameFormat();
  tf.setMargin(8);
  m_doc.rootFrame()->setFrameFormat(tf);
  update();
}

void Preview::paintEvent(QPaintEvent*)
{
  QPainter painter(this);
  double f = height() / 40;
  painter.scale(f, f);
  m_doc.drawContents(&painter);
}

EditTextDlg::EditTextDlg(QWidget* parent, QLineEdit* e) :
 QDialog(parent), sfont("#NN"), ed(e)
{
  setWindowTitle(i18n("Edit text"));
  QVBoxLayout* vbox = new QVBoxLayout(this);
  vbox->addWidget(editPreview = new Preview);
  editPreview->setMinimumHeight(40);
  editPreview->setText(ed->text());
  QGridLayout* grid = new QGridLayout;
  vbox->addLayout(grid);
  grid->addWidget(text = new QLineEdit(ed->text()), 0, 0);
  text->setCursorPosition(ed->text().size());
  connect(text, SIGNAL(textChanged(const QString&)),
          editPreview, SLOT(setText(const QString&)));
  connect(text, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  QHBoxLayout* hbox = new QHBoxLayout;
  grid->addLayout(hbox, 0, 1);
  bCat = new QToolButton;
  bCat->setArrowType(Qt::LeftArrow);
  hbox->addWidget(bCat);
  connect(bCat, SIGNAL(clicked()), SLOT(slotCat()));
  hbox->addWidget(font = new QLineEdit(sfont));
  font->setFixedWidth(60);
  QGroupBox* g = new QGroupBox;
  grid->addWidget(g, 1, 0);
  g->setTitle(i18n("Font"));
  vbox = new QVBoxLayout(g);
  QButtonGroup* bg = new QButtonGroup;
  QRadioButton* rb;
  int i = 0;
  vbox->addWidget(rb = new QRadioButton("Helvetica"));
  bg->addButton(rb, i++);
  rb->setChecked(true);
  vbox->addWidget(rb = new QRadioButton("Helvetica Bold"));
  bg->addButton(rb, i++);
  vbox->addWidget(rb = new QRadioButton("Helvetica Italic"));
  bg->addButton(rb, i++);
  vbox->addWidget(rb = new QRadioButton("Helvetica Bold Italic"));
  bg->addButton(rb, i++);
  vbox->addWidget(rb = new QRadioButton("Symbol"));
  bg->addButton(rb, i++);
  vbox->addWidget(rb = new QRadioButton("Dingbats"));
  bg->addButton(rb, i++);
  connect(bg, SIGNAL(buttonClicked(int)), SLOT(slotFont(int)));
  grid->addLayout(vbox = new QVBoxLayout, 1, 1);
  vbox->addWidget(g = new QGroupBox);
  g->setTitle(i18n("Position"));
  QVBoxLayout* vbox2 = new QVBoxLayout(g);
  bg = new QButtonGroup;
  i = 0;
  vbox2->addWidget(rb = new QRadioButton(i18n("Up")));
  bg->addButton(rb, i++);
  vbox2->addWidget(rb = new QRadioButton(i18n("Normal")));
  bg->addButton(rb, i++);
  rb->setChecked(true);
  vbox2->addWidget(rb = new QRadioButton(i18n("Down")));
  bg->addButton(rb, i++);
  connect(bg, SIGNAL(buttonClicked(int)), SLOT(slotPos(int)));
  vbox->addWidget(buttonBox = new QDialogButtonBox(QDialogButtonBox::Help |
                                                   QDialogButtonBox::Ok |
                                                   QDialogButtonBox::Cancel));
  buttonBox->setOrientation(Qt::Vertical);
  QPushButton* bOk = buttonBox->button(QDialogButtonBox::Ok);
  bOk->setDefault(true);
  bOk->setEnabled(false);
  connect(buttonBox, SIGNAL(clicked(QAbstractButton*)),
          SLOT(slotButtonClicked(QAbstractButton*)));
  text->setFocus();
  QTimer::singleShot(0, this, SLOT(slotRestoreSize()));
}

EditTextDlg::~EditTextDlg()
{
  KConfigGroup config = KSharedConfig::openConfig()->group("EditTextDlg");
  KWindowConfig::saveWindowSize(windowHandle(), config);
}

void EditTextDlg::editText(QWidget* parent, QLineEdit* ed)
{
  EditTextDlg dlg(parent, ed);
  dlg.exec();
}

void EditTextDlg::keyPressEvent(QKeyEvent* e)
{
  if ((e->QInputEvent::modifiers() == Qt::NoModifier) &&
      (e->key() == Qt::Key_F1)) {
    buttonBox->button(QDialogButtonBox::Help)->animateClick();
    e->accept();
  } else
    QDialog::keyPressEvent(e);
}

void EditTextDlg::slotButtonClicked(QAbstractButton* button)
{
  int id = buttonBox->standardButton(button);
  switch (id) {
    case QDialogButtonBox::Ok: {
      ed->setText(text->text());
      accept();
      break;
    }
    case QDialogButtonBox::Help:
      QDesktopServices::openUrl(QUrl(QStringLiteral("help:/kpl/"
                                                    "edit-text.html")));
      break;
    case QDialogButtonBox::Cancel:
      reject();
  }
}

void EditTextDlg::slotCat()
{
  text->insert(font->text());
  text->setFocus();
}

void EditTextDlg::slotEdited()
{
  buttonBox->button(QDialogButtonBox::Ok)->setEnabled(true);
}

void EditTextDlg::slotFont(int id)
{
  const char fs[] = "HBOISZ";
  sfont[1] = fs[id];
  font->setText(sfont);
  bCat->setFocus();
}

void EditTextDlg::slotPos(int id)
{
  const char ps[] = "UND";
  sfont[2] = ps[id];
  font->setText(sfont);
  bCat->setFocus();
}

void EditTextDlg::slotRestoreSize()
{
  KWindowConfig::restoreWindowSize(windowHandle(),
    KSharedConfig::openConfig()->group("EditTextDlg"));
}
