/***************************************************************************
                          funcdlg.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Apr 25 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FUNCDLG_H
#define FUNCDLG_H

#include <QDialog>

class QDialogButtonBox;
class QLineEdit;
class KColorButton;
class KUrlRequester;
class BrushButton;
class DoubleEdit;
class DoubleSpinBox;
class FunItem;
class KplDoc;
class ParFunItem;
class SpinBox;
class SymbolButton;

/**
  * Function settings dialog class. Provides a dialog for the settings
  * of an ordinary function.
  * @author Werner Stille
  */
class FuncDlg : public QDialog
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param parent pointer to parent widget.
   * @param model pointer to KplDoc model object.
   * @param fi pointer to function object.
   * @param pos dialog position.
   */
  FuncDlg(QWidget* parent, KplDoc* model, FunItem* fi, const QPoint& pos);
  /** Destructor. */
  ~FuncDlg();

protected:
  /**
   * Applies settings.
   * @param ok true to close the dialog.
   */
  void apply(bool ok = false);
  /**
   * Copies settings to temporary function object.
   * @return false for error.
   */
  bool getTemp();
  /**
   * Reimplemented to display help when key F1 is pressed.
   * @param e Key event.
   */
  virtual void keyPressEvent(QKeyEvent* e);
  KplDoc* m;
  FunItem *fi, *fit;
  KUrlRequester* fileName;
  QLineEdit* func;
  QDialogButtonBox* buttonBox;
  BrushButton* fill;
  SymbolButton* bSymb;
  DoubleEdit *exMin, *exMax, *exStep, *efx, *efy;
  DoubleSpinBox *ex0, *ey0;
  SpinBox *symb, *eRelSize;
  KColorButton* colData;

protected slots:
  /** Applies settings without leaving the dialog. */
  void slotApply();
  /** Reimplemented from QDialog. */
  virtual void slotButtonClicked(QAbstractButton* button);
  /** Enables buttons. */
  void slotEdited();
  /** Displays the dialog for parameter editing. */
  void slotEditParameter();
  /** Exports to file. */
  void slotExport();
  /**
   * Adjusts state of widgets when fill pattern is changed.
   * @param brush new pattern number.
   */
  void slotFillChanged(int brush);
  /** Displays the dialog for selection of a function. */
  void slotFunc();
  /** Displays the dialog for selection of a parameter file to read. */
  void slotGetParameter();
  /** Restores window size. */
  void slotRestoreSize();
  /** Caculates roots. */
  void slotRoots();
  /** Displays the dialog for selection of a parameter file to save. */
  void slotSaveParameter();
};

/**
  * Parametric function settings dialog class. Provides a dialog for the settings
  * of parametric functions.
  * @author Werner Stille
  */
class ParFuncDlg : public QDialog
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param parent pointer to parent widget.
   * @param model pointer to KplDoc model object.
   * @param fi pointer to parametric function object.
   * @param pos dialog position.
   */
  ParFuncDlg(QWidget* parent, KplDoc* model, ParFunItem* fi,
             const QPoint& pos);
  /** Destructor. */
  ~ParFuncDlg();

protected:
  /**
   * Applies settings.
   * @param ok true to close the dialog.
   */
  void apply(bool ok = false);
  /**
   * Copies settings to temporary parametric function object.
   * @return false for error.
   */
  bool getTemp();
  /**
   * Reimplemented to display help when key F1 is pressed.
   * @param e Key event.
   */
  virtual void keyPressEvent(QKeyEvent* e);
  KplDoc* m;
  ParFunItem *fi, *fit;
  KUrlRequester *fileNameX, *fileNameY;
  QDialogButtonBox* buttonBox;
  QLineEdit *funcX, *funcY;
  BrushButton* fill;
  SymbolButton* bSymb;
  DoubleEdit *etMin, *etMax, *etStep, *efx, *efy;
  DoubleSpinBox *ex0, *ey0;
  SpinBox *symb, *eRelSize;
  KColorButton* colData;

protected slots:
  /** Applies settings without leaving the dialog. */
  void slotApply();
  /**
   * Handles button clicks.
   * @param button pointer to button.
   */
  virtual void slotButtonClicked(QAbstractButton* button);
  /** Enables buttons. */
  void slotEdited();
  /** Displays the dialog for parameter editing for the x axis function. */
  void slotEditParameterX();
  /** Displays the dialog for parameter editing for the y axis function. */
  void slotEditParameterY();
  /** Exports to file. */
  void slotExport();
  /**
   * Adjusts state of widgets when fill pattern is changed.
   * @param brush new pattern number.
   */
  void slotFillChanged(int brush);
  /** Displays the dialog for selection of a function for the x axis. */
  void slotFuncX();
  /** Displays the dialog for selection of a function for the y axis. */
  void slotFuncY();
  /** Displays the dialog for selection of a parameter file to read for the x axis function. */
  void slotGetParameterX();
  /** Displays the dialog for selection of a parameter file to read for the y axis function. */
  void slotGetParameterY();
  /** Restores window size. */
  void slotRestoreSize();
  /** Displays the dialog for selection of a parameter file to save for the x axis function. */
  void slotSaveParameterX();
  /** Displays the dialog for selection of a parameter file to save for the y axis function. */
  void slotSaveParameterY();
};

#endif
