/***************************************************************************
                          imagedlg.cpp  -  description
                             -------------------
    begin                : Don Jun 5 2003
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QCheckBox>
#include <QDesktopServices>
#include <QDialogButtonBox>
#include <QDoubleSpinBox>
#include <QKeyEvent>
#include <QLabel>
#include <QLayout>
#include <QTimer>
#include <QUndoStack>
#include <KLocalizedString>
#include <KUrlRequester>
#include <KMessageBox>
#include <KSharedConfig>
#include <KWindowConfig>
#include "imagedlg.h"
#include "imageitem.h"
#include "kpldoc.h"
#include "kplsettings.h"
#include "utils.h"

ImageDlg::ImageDlg(QWidget* parent, KplDoc* model, ImageItem* ii0,
                   const QPoint& pos) :
 QDialog(parent), m(model), ii(ii0)
{
  setWindowTitle(i18n("Image"));
  QVBoxLayout* vbox = new QVBoxLayout(this);
  QGridLayout* grid = new QGridLayout;
  vbox->addLayout(grid);
  grid->addWidget(new QLabel(i18n("File")), 0, 0);
  grid->addWidget(fileName = new KUrlRequester(ii->url()), 0, 1, 1, 4);
  fileName->setStartDir(Utils::url(KplSettings::currentDir()));
  fileName->setFilter("*.eps *.EPS *.epsi *.EPSI *.epsf *.EPSF\n*");
  grid->addWidget(new QLabel(i18n("Rotation")), 1, 0);
  grid->addWidget(eDir = new QDoubleSpinBox, 1, 1);
  eDir->setRange(-360, 360);
  eDir->setSuffix(QString::fromUtf8("\302\260"));
  eDir->setValue(ii->dir());
  grid->addWidget(new QLabel("x0"), 2, 0);
  grid->addWidget(ex0 = new QDoubleSpinBox, 2, 1);
  ex0->setRange(-200, 200);
  ex0->setSingleStep(0.1);
  ex0->setSuffix(" cm");
  ex0->setValue(ii->x0());
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 2, 2);
  grid->addWidget(new QLabel("y0"), 2, 3);
  grid->addWidget(ey0 = new QDoubleSpinBox, 2, 4);
  ey0->setRange(-200, 200);
  ey0->setSingleStep(0.1);
  ey0->setSuffix(" cm");
  ey0->setValue(ii->y0());
  grid->addWidget(new QLabel(i18n("x scaling")), 3, 0);
  grid->addWidget(efx = new QDoubleSpinBox, 3, 1);
  efx->setDecimals(1);
  efx->setRange(0.1, 800);
  efx->setSuffix(" %");
  efx->setValue(100 * ii->scaleX());
  connect(efx, SIGNAL(valueChanged(const QString&)), SLOT(slotAspect()));
  grid->addWidget(new QLabel(i18n("y scaling")), 3, 3);
  grid->addWidget(efy = new QDoubleSpinBox, 3, 4);
  efy->setDecimals(1);
  efy->setRange(0.1, 800);
  efy->setSuffix(" %");
  efy->setValue(100 * ii->scaleY());
  QHBoxLayout* hbox;
  grid->addLayout(hbox = new QHBoxLayout, 4, 1, 1, 4);
  hbox->addWidget(isotropic = new QCheckBox(i18n("Maintain aspect ratio")));
  hbox->addItem(new QSpacerItem(10, 10, QSizePolicy::Expanding));
  connect(isotropic, SIGNAL(toggled(bool)), SLOT(slotAspect()));
  isotropic->setChecked(ii->scaleX() == ii->scaleY());
  vbox->addWidget(buttonBox = new QDialogButtonBox(QDialogButtonBox::Help |
                                                   QDialogButtonBox::Ok |
                                                   QDialogButtonBox::Apply |
                                                   QDialogButtonBox::Cancel));
  buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
  move(pos);
  buttonBox->button(QDialogButtonBox::Apply)->setEnabled(false);
  connect(fileName, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(ex0, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  connect(ey0, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  connect(eDir, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  connect(efx, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  connect(efy, SIGNAL(valueChanged(const QString&)), SLOT(slotEdited()));
  connect(buttonBox, SIGNAL(clicked(QAbstractButton*)),
          SLOT(slotButtonClicked(QAbstractButton*)));
  QTimer::singleShot(0, this, SLOT(slotRestoreSize()));
}

ImageDlg::~ImageDlg()
{
  KConfigGroup config = KSharedConfig::openConfig()->group("ImageDlg");
  KWindowConfig::saveWindowSize(windowHandle(), config);
}

void ImageDlg::keyPressEvent(QKeyEvent* e)
{
  if ((e->QInputEvent::modifiers() == Qt::NoModifier) &&
      (e->key() == Qt::Key_F1)) {
    buttonBox->button(QDialogButtonBox::Help)->animateClick();
    e->accept();
  } else
    QDialog::keyPressEvent(e);
}

void ImageDlg::slotAspect()
{
  bool iso = isotropic->isChecked();
  if (iso)
    efy->setValue(efx->value());
  efy->setDisabled(iso);
}

void ImageDlg::slotButtonClicked(QAbstractButton* button)
{
  int id = buttonBox->standardButton(button);
  switch (id) {
    case QDialogButtonBox::Apply:
      buttonBox->button(QDialogButtonBox::Apply)->setEnabled(false);
      buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
    case QDialogButtonBox::Ok: {
      QUrl u = fileName->url();
      if (ImageItem::checkBoundingBox(u)) {
        ImageItem* iiNew = new ImageItem(*ii);
        iiNew->setUrl(u);
        iiNew->setDir(eDir->value());
        iiNew->setX0(ex0->value());
        iiNew->setY0(ey0->value());
        double fx = 0.01 * efx->value();
        iiNew->setScaleX(fx);
        iiNew->setScaleY(isotropic->isChecked() ? fx : 0.01 * efy->value());
        m->undoStack()->push(new EditCommand(iiNew, ii, m));
        ii = iiNew;
      } else {
        KMessageBox::sorry(this, i18n("not an EPS file!"));
        return;
      }
      if (id == QDialogButtonBox::Ok)
        accept();
      break;
    }
    case QDialogButtonBox::Help:
      QDesktopServices::openUrl(QUrl(QStringLiteral("help:/kpl/image.html")));
      break;
    case QDialogButtonBox::Cancel:
      reject();
  }
}

void ImageDlg::slotEdited()
{
  buttonBox->button(QDialogButtonBox::Apply)->setEnabled(true);
  buttonBox->button(QDialogButtonBox::Ok)->setEnabled(true);
}

void ImageDlg::slotRestoreSize()
{
  KWindowConfig::restoreWindowSize(windowHandle(),
    KSharedConfig::openConfig()->group("ImageDlg"));
}
