/***************************************************************************
                          kpl.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Apr 24 15:14:00 MEST 1999

    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QApplication>
#include <QClipboard>
#include <QDockWidget>
#include <QFileInfo>
#include <QHeaderView>
#include <QInputDialog>
#include <QLabel>
#include <QMenu>
#include <QMimeData>
#include <QMouseEvent>
#include <QPrintDialog>
#include <QPrinter>
#include <QStatusBar>
#include <QThread>
#include <QTimer>
#include <QTreeView>
#include <KActionCollection>
#include <KConfigDialog>
#include <KConfigDialogManager>
#include <KConfigGroup>
#include <KLocalizedString>
#include <KMessageBox>
#include <KNotifyConfigWidget>
#include <KRecentFilesAction>
#include <KSharedConfig>
#include <KToggleAction>
#include <KUndoActions>
#include <KXMLGUIFactory>
#include <ThreadWeaver/ThreadWeaver>
#include "arcitem.h"
#include "arrayitem.h"
#include "array3ditem.h"
#include "arrowitem.h"
#include "ellipseitem.h"
#include "filedialog.h"
#include "fitdlg.h"
#include "frameitem.h"
#include "fun3ditem.h"
#include "imageitem.h"
#include "kpl.h"
#include "kpladaptor.h"
#include "kpldoc.h"
#include "kplsettings.h"
#include "kplutils.h"
#include "kplview.h"
#include "legenditem.h"
#include "lineitem.h"
#include "newitemdlg.h"
#include "parfunitem.h"
#include "rectitem.h"
#include "refreshdlg.h"
#include "scalebaritem.h"
#include "settingsdlg.h"
#include "splfitdlg.h"
#include "splineitem.h"
#include "textitem.h"
#include "utils.h"

static int precision(double x, double dx)
{
  return qBound(2, int(ceil(log10(qMax(fabs(x), 1e-300)))
                       - floor(log10(qMax(fabs(dx), 1e-300)))), 16);
}

Kpl::Kpl() :
 reloadActive(false), xPrec(4), yPrec(4), doc(new KplDoc(this)),
 timer(new QTimer(this)), statusMsgLabel(new QLabel()),
 statusAddLabel(new QLabel()), statusItemLabel(new QLabel()),
 statusPosLabel(new QLabel()), statusValLabel(new QLabel())
{
  new KplAdaptor(this);
  QDBusConnection dbus = QDBusConnection::sessionBus();
  dbus.registerObject("/Kpl", this);
  dbus.registerService("org.kde.Kpl");
  timer->setInterval(5000);
  KConfigDialogManager::changedMap()->insert("DoubleEdit",
                                             SIGNAL(changed(double)));
  KConfigDialogManager::changedMap()->insert("BrushButton",
                                             SIGNAL(changed(int)));
  itemWidget = new QDockWidget(i18n("Items"), this);
  itemWidget->setWidget(itemList = new QTreeView);
  itemWidget->setObjectName("Items");
  itemList->setModel(doc);
  itemList->setAlternatingRowColors(true);
  itemList->setSelectionMode(QAbstractItemView::ExtendedSelection);
  itemList->setDragEnabled(true);
  itemList->setDropIndicatorShown(true);
  itemList->setDragDropMode(QAbstractItemView::DragDrop);
  itemList->setContextMenuPolicy(Qt::CustomContextMenu);
  itemList->header()->setSectionResizeMode(QHeaderView::ResizeToContents);
  itemList->header()->setSectionsMovable(false);
  connect(itemList, SIGNAL(customContextMenuRequested(const QPoint&)),
          SLOT(slotPopup(const QPoint&)));
  addDockWidget(Qt::LeftDockWidgetArea, itemWidget);
  QAction* a = KStandardAction::openNew(this, SLOT(slotFileNew()),
                                        actionCollection());
  QIcon icon  = QIcon::fromTheme(QStringLiteral("kpl-document-open"));
  if (icon.isNull())
    icon  = QIcon::fromTheme(QStringLiteral("kpl-plot"));
  a->setStatusTip(i18n("Creates a new plot"));
  a = new QAction(icon, i18n("&Open Plot File..."), actionCollection());
  actionCollection()->addAction("open_plot", a);
  actionCollection()->setDefaultShortcut(a, Qt::CTRL + Qt::Key_O);
  a->setStatusTip(i18n("Opens an existing plot file"));
  connect(a, SIGNAL(triggered()), SLOT(slotFileOpenPlot()));
  a = new QAction(QIcon::fromTheme(QStringLiteral("document-open")),
                  i18n("Open &Data File..."), actionCollection());
  actionCollection()->addAction("open_data", a);
  a->setStatusTip(i18n("Opens an existing data file"));
  connect(a, SIGNAL(triggered()), SLOT(slotFileOpenData()));
  recentFiles = KStandardAction::openRecent(this,
                                            SLOT(slotOpenRecent(const QUrl&)),
                                            actionCollection());
  recentFiles->setStatusTip(i18n("Opens a recently used file"));
  recentFiles->loadEntries(KSharedConfig::openConfig()->group("Recent Files"));
  fileSave = KStandardAction::save(this, SLOT(slotFileSave()),
                                   actionCollection());
  fileSave->setStatusTip(i18n("Saves the current plot"));
  fileSaveAs = KStandardAction::saveAs(this, SLOT(slotFileSaveAs()),
                                       actionCollection());
  fileSaveAs->setStatusTip(i18n("Saves the plot under a different name"));
  filePrint = KStandardAction::print(this, SLOT(slotFilePrint()),
                                     actionCollection());
  filePrint->setStatusTip(i18n("Prints the current plot"));
  filePSP = new QAction(QIcon::fromTheme(QStringLiteral("document-export")),
                        i18n("&Portrait..."), actionCollection());
  actionCollection()->addAction("postscript_p", filePSP);
  connect(filePSP, SIGNAL(triggered()), SLOT(slotFilePSP()));
  filePSP->setStatusTip(i18n("Generates PostScript file (portrait)"));
  filePSL = new QAction(QIcon::fromTheme(QStringLiteral("document-export")),
                        i18n("&Landscape..."), actionCollection());
  actionCollection()->addAction("postscript_l", filePSL);
  connect(filePSL, SIGNAL(triggered()), SLOT(slotFilePSL()));
  filePSL->setStatusTip(i18n("Generates PostScript file (landscape)"));
  filePreviewP =
    new QAction(QIcon::fromTheme(QStringLiteral("document-preview")),
                i18n("&Portrait..."), actionCollection());
  actionCollection()->addAction("preview_p", filePreviewP);
  connect(filePreviewP, SIGNAL(triggered()), SLOT(slotFilePreviewP()));
  filePreviewP->setStatusTip(i18n("Displays PostScript preview (portrait)"));
  filePreviewL =
    new QAction(QIcon::fromTheme(QStringLiteral("document-preview")),
                i18n("&Landscape..."), actionCollection());
  actionCollection()->addAction("preview_l", filePreviewL);
  connect(filePreviewL, SIGNAL(triggered()), SLOT(slotFilePreviewL()));
  filePreviewP->setStatusTip(i18n("Displays PostScript preview (landscape)"));
  fileExport = new QAction(QIcon::fromTheme(QStringLiteral("document-export")),
                           i18n("&Export..."), actionCollection());
  actionCollection()->addAction("export", fileExport);
  connect(fileExport, SIGNAL(triggered()), SLOT(slotFileExport()));
  fileExport->setStatusTip(i18n("Exports to image file"));
  a = new QAction(windowIcon(), i18n("New &Window"),
                  actionCollection());
  actionCollection()->addAction("new_window", a);
  a->setStatusTip(i18n("Opens a new application window"));
  connect(a, SIGNAL(triggered()), SLOT(slotFileNewWindow()));
  a = new QAction(QIcon::fromTheme(QStringLiteral("window-close")),
                  i18n("Close W&indow"), actionCollection());
  actionCollection()->addAction("close_window", a);
  a->setStatusTip(i18n("Closes the current window"));
  connect(a, SIGNAL(triggered()), SLOT(close()));
  a = KStandardAction::quit(this, SLOT(slotFileQuit()), actionCollection());
  a->setStatusTip(i18n("Exits the program"));
  a = KUndoActions::createUndoAction(doc->undoStack(),actionCollection());
// TODO: Undo action text is sometimes not translated.
  QString s = i18n("Undo");
  if (s != a->text())
    a->setText(s);
  a = KUndoActions::createRedoAction(doc->undoStack(),actionCollection());
// TODO: Redo action text is sometimes not translated.
  s = i18n("Redo");
  if (s != a->text())
    a->setText(s);
  a = new QAction(QIcon::fromTheme(QStringLiteral("archive-insert")),
                  i18n("&New Item..."), actionCollection());
  actionCollection()->addAction("edit_new", a);
  a->setStatusTip(i18n("Inserts new item"));
  connect(a, SIGNAL(triggered()), SLOT(slotNewItem()));
  editItem = new QAction(QIcon::fromTheme(QStringLiteral("document-edit")),
                         i18n("&Edit..."), actionCollection());
  actionCollection()->addAction("edit_item", editItem);
  actionCollection()->setDefaultShortcut(editItem, Qt::Key_F2);
  editItem->setStatusTip(i18n("Edits item"));
  connect(editItem, SIGNAL(triggered()), SLOT(slotEdit()));
  icon  = QIcon::fromTheme(QStringLiteral("kpl-fit"));
  if (icon.isNull())
    icon  = QIcon::fromTheme(QStringLiteral("configure"));
  editFit = new QAction(icon, i18n("&Fit..."),  actionCollection());
  actionCollection()->addAction("edit_fit", editFit);
  editFit->setStatusTip(i18n("Fits items"));
  connect(editFit, SIGNAL(triggered()), SLOT(slotFit()));
  editDelete = new QAction(QIcon::fromTheme(QStringLiteral("archive-remove")),
                           i18n("&Delete"), actionCollection());
  actionCollection()->addAction("edit_delete", editDelete);
  actionCollection()->setDefaultShortcut(editDelete, Qt::Key_Delete);
  editDelete->setStatusTip(i18n("Deletes item"));
  connect(editDelete, SIGNAL(triggered()), SLOT(slotDelete()));
  editCopy = KStandardAction::copy(this, SLOT(slotCopy()), actionCollection());
  editCopy->setStatusTip(i18n("Copies selected items to the clipboard"));
  editCut = KStandardAction::cut(this, SLOT(slotCut()), actionCollection());
  editCut->setStatusTip(i18n("Moves selected items to the clipboard"));
  editPaste = KStandardAction::paste(this, SLOT(slotPaste()),
                                     actionCollection());
  editPaste->setStatusTip(i18n("Pastes items"));
  editActive = new KToggleAction(i18n("&Active"), actionCollection());
  actionCollection()->addAction("edit_active", editActive);
  editActive->setStatusTip(i18n("Toggles visibility of item"));
  connect(editActive, SIGNAL(triggered()), SLOT(slotActive()));
  editUp = new QAction(QIcon::fromTheme(QStringLiteral("go-up")), i18n("&Up"),
                       actionCollection());
  actionCollection()->addAction("edit_up", editUp);
  actionCollection()->setDefaultShortcut(editUp, Qt::CTRL + Qt::Key_Up);
  editUp->setStatusTip(i18n("Moves item upwards"));
  connect(editUp, SIGNAL(triggered()), SLOT(slotEditUp()));
  editDown = new QAction(QIcon::fromTheme(QStringLiteral("go-down")),
                         i18n("&Down"), actionCollection());
  actionCollection()->addAction("edit_down", editDown);
  actionCollection()->setDefaultShortcut(editDown, Qt::CTRL + Qt::Key_Down);
  editDown->setStatusTip(i18n("Moves item downwards"));
  connect(editDown, SIGNAL(triggered()), SLOT(slotEditDown()));
  editTop = new QAction(QIcon::fromTheme(QStringLiteral("go-top")),
                        i18n("&Top"), actionCollection());
  actionCollection()->addAction("edit_top", editTop);
  editTop->setStatusTip(i18n("Moves item to top"));
  connect(editTop, SIGNAL(triggered()), SLOT(slotEditTop()));
  editBottom = new QAction(QIcon::fromTheme(QStringLiteral("go-bottom")),
                           i18n("&Bottom"), actionCollection());
  actionCollection()->addAction("edit_bottom", editBottom);
  editBottom->setStatusTip(i18n("Moves item to bottom"));
  connect(editBottom, SIGNAL(triggered()), SLOT(slotEditBottom()));
  a = new QAction(QIcon::fromTheme(QStringLiteral("format-justify-fill")),
                  i18n("&Items..."), actionCollection());
  actionCollection()->addAction("edit_items", a);
  a->setStatusTip(i18n("Edits plot items"));
  connect(a, SIGNAL(triggered()), SLOT(slotEditItems()));
  a = KStandardAction::zoomIn(this, SLOT(slotViewZoomIn()),
                              actionCollection());
  a->setStatusTip(i18n("Increases magnification by 20 %"));
  a = KStandardAction::zoomOut(this, SLOT(slotViewZoomOut()),
                               actionCollection());
  a->setStatusTip(i18n("Decreases magnification by 20 %"));
  a = KStandardAction::zoom(this, SLOT(slotViewZoom()), actionCollection());
  a->setStatusTip(i18n("Sets magnification"));
  viewReload = KStandardAction::redisplay(this, SLOT(slotViewReload()),
                                          actionCollection());
  viewReload->setStatusTip(i18n("Reloads file"));
  a = new QAction(QIcon::fromTheme(QStringLiteral("download-later")),
                  i18n("Reload &Periodically..."), actionCollection());
  actionCollection()->addAction("autoreload", a);
  a->setStatusTip(i18n("Reloads file periodically"));
  connect(a, SIGNAL(triggered()), SLOT(slotViewAutoReload()));
  showSource = new KToggleAction(i18n("Show Function Sou&rce"),
                                 actionCollection());
  actionCollection()->addAction("show_source", showSource);
  showSource->setStatusTip(i18n("Enables / disables display of "
                                "function source"));
  connect(showSource, SIGNAL(triggered()), SLOT(slotOptionsShowSource()));
  addItems = new KToggleAction(i18n("Add &Files"), actionCollection());
  actionCollection()->addAction("add_items", addItems);
  actionCollection()->setDefaultShortcut(addItems, Qt::Key_Insert);
  addItems->setStatusTip(i18n("Enables / disables adding of data and "
                              "plot files"));
  connect(addItems, SIGNAL(triggered()), SLOT(slotOptionsAddData()));
  boundingBox = new KToggleAction(i18n("Ca&lculate Bounding Box"),
                                  actionCollection());
  actionCollection()->addAction("bounding_box", boundingBox);
  boundingBox->setStatusTip(i18n("Enables / disables calculation of "
                                 "bounding box"));
  connect(boundingBox, SIGNAL(triggered()), SLOT(slotOptionsBoundingBox()));
  saveAbsPath = new KToggleAction(i18n("Save Absol&ute Paths"),
                                  actionCollection());
  actionCollection()->addAction("save_abs_path", saveAbsPath);
  saveAbsPath->setStatusTip(i18n("Saves absolute paths of data "
                                 "and library files in plot files"));
  connect(saveAbsPath, SIGNAL(triggered()), SLOT(slotOptionsSaveAbsPath()));
  warning = new KToggleAction(i18n("Unsaved Changes &Warning"),
                              actionCollection());
  actionCollection()->addAction("warning", warning);
  warning->setStatusTip(i18n("Enables / disables warning on "
                           "unsaved changes"));
  connect(warning, SIGNAL(triggered()), SLOT(slotOptionsWarning()));
  a = KStandardAction::configureNotifications(this,
                                              SLOT(slotOptionsNotifications()),
                                              actionCollection());
  a->setStatusTip(i18n("Displays notification configuration dialog"));
  a = KStandardAction::preferences(this, SLOT(slotOptionsPreferences()),
                                   actionCollection());
  a->setStatusTip(i18n("Displays the configuration dialog of Kpl"));
  createStandardStatusBarAction();
  setupGUI();
//   setupGUI(Default, "/home/stille/projects/kpl/kpl50/src/kplui.rc");
  a = actionCollection()->action(
    KStandardAction::name(KStandardAction::ShowStatusbar));
  if (a)
    a->setStatusTip(i18n("Enables/disables the statusbar"));
  actionCollection()->action(
    KStandardAction::name(KStandardAction::KeyBindings))->
      setStatusTip(i18n("Displays key bindings dialog"));
  actionCollection()->action(
    KStandardAction::name(KStandardAction::ConfigureToolbars))->
      setStatusTip(i18n("Displays toolbar configuration dialog"));
  actionCollection()->removeAction(actionCollection()->action(
    KStandardAction::name(KStandardAction::WhatsThis)));
  actionCollection()->action(
    KStandardAction::name(KStandardAction::HelpContents))->
      setStatusTip(i18n("Displays the Kpl handbook"));
  actionCollection()->action(
    KStandardAction::name(KStandardAction::ReportBug))->
      setStatusTip(i18n("Displays a dialog for submitting a bug report"));
  actionCollection()->action(
    KStandardAction::name(KStandardAction::SwitchApplicationLanguage))->
      setStatusTip(i18n("Switches the language of the application"));
  actionCollection()->action(
    KStandardAction::name(KStandardAction::AboutApp))->
      setStatusTip(i18n("Displays essential information about Kpl"));
  actionCollection()->action(
    KStandardAction::name(KStandardAction::AboutKDE))->
      setStatusTip(i18n("Displays essential information about KDE"));
  toolBarMenuAction()->setStatusTip(i18n("Enables/disables the toolbar"));
  doc->newDocument();
  view = new KplView(KplSettings::zoom());
  view->setScene(doc->scene());
  setCentralWidget(view);
  initStatusBar();
  updWin();
  settingsChanged();
  connect(doc, SIGNAL(expanded(const QModelIndex&)), itemList,
          SLOT(expand(const QModelIndex&)));
  connect(view, SIGNAL(regionSelected(const QRectF&)), doc,
          SLOT(zoom(const QRectF&)));
  connect(itemList, SIGNAL(expanded(const QModelIndex&)), doc,
          SLOT(expand(const QModelIndex&)));
  connect(itemList, SIGNAL(collapsed(const QModelIndex&)), doc,
          SLOT(collapse(const QModelIndex&)));
  connect(doc->scene(), SIGNAL(mouseMoved(const QPointF&, const FrameItem*)),
          SLOT(updStatus(const QPointF&, const FrameItem*)));
  connect(doc->scene(), SIGNAL(selectionChanged()), SLOT(select()));
  connect(doc->scene(), SIGNAL(previewCreated(const QString&)),
          SLOT(addPreview(const QString&)));
  connect(view, SIGNAL(mouseLeaved()), SLOT(clearStatus()));
  connect(view, SIGNAL(customContextMenuRequested(const QPoint&)),
          SLOT(slotPopupItem(const QPoint&)));
  connect(doc, SIGNAL(modelReset()), SLOT(updWin()));
  connect(itemList->selectionModel(),
          SIGNAL(selectionChanged(const QItemSelection&,
                 const QItemSelection&)), SLOT(updWin()));
  connect(timer, SIGNAL(timeout()), SLOT(slotWatchFile()));
  connect(doc, SIGNAL(displayMessage(const QString&)),
          SLOT(statusMsg(const QString&)));
  connect(view, SIGNAL(urlsDropped(QList<QUrl>*)),
          SLOT(slotOpenList(QList<QUrl>*)));
  connect((QMenu*) factory()->container("edit", this), SIGNAL(aboutToShow()),
          SLOT(updatePasteAction()));
  connect((QMenu*) factory()->container("popup_item", this),
          SIGNAL(aboutToShow()), SLOT(updatePasteAction()));
  connect((QMenu*) factory()->container("popup", this), SIGNAL(aboutToShow()),
          SLOT(updatePasteAction()));
}

Kpl::~Kpl()
{
  for (int i = 0; i < previews.count(); i++)
    QFile::remove(previews[i]);
}

void Kpl::openList(QList<QUrl>* list, ArrayItem::DecimalCharacter idec, int type)
{
  statusMsg(i18n("Opening file..."));
  if (idec == ArrayItem::Settings)
    idec = KplUtils::decimalCharacter();
  for (int i = 0; i < list->count(); i++)
    if ((!i) || KplSettings::addItems()) {
      if (KplSettings::addItems() || (KplSettings::warning() ? queryClose() :
                                                               true))
        doc->openDocument((*list)[i], idec, KplDoc::FileType(type));
    } else {
      Kpl* kpl = new Kpl;
      if (kpl->doc->openDocument((*list)[i], idec, KplDoc::FileType(type)))
        kpl->show();
      else
        delete kpl;
    }
  ready();
}

void Kpl::slotOpenList(QList<QUrl>* list)
{
  openList(list, KplUtils::decimalCharacter(), 0);
}

void Kpl::newPlot()
{
  doc->newDocument();
}

bool Kpl::openPlot(const QString& url)
{
  return doc->openDocument(Utils::url(url), ArrayItem::Point, KplDoc::Plot);
}

bool Kpl::openData(const QString& url)
{
  return doc->openDocument(Utils::url(url), KplUtils::decimalCharacter(),
                           KplDoc::Data);
}

bool Kpl::savePlot(const QString& url)
{
  if (doc->rowCount())
    return saveAs(Utils::url(url));
  else
    return false;
}

bool Kpl::savePS(const QString& url, bool landscape)
{
  if (doc->rowCount())
    return savePSUrl(Utils::url(url), landscape);
  else
    return false;
}

bool Kpl::exportImage(const QString& url)
{
  if (doc->rowCount())
    return exportImageUrl(Utils::url(url));
  else
    return false;
}

bool Kpl::printPlot(const QString& printerName, const QString& fileName,
                    bool landscape, int numCopies)
{
  if (doc->rowCount()) {
    QPrinter p(QPrinter::HighResolution);
    p.setColorMode(QPrinter::Color);
    p.setPrinterName(printerName);
    p.setOutputFileName(fileName);
    p.setOrientation(landscape ? QPrinter::Landscape : QPrinter::Portrait);
    p.setNumCopies(numCopies);
    doc->scene()->print(&p);
    return true;
  } else
    return false;
}

void Kpl::setAddFilesOption(bool add)
{
  addItems->setChecked(add);
  slotOptionsAddData();
}

void Kpl::setBoundingBoxOption(bool calculate)
{
  boundingBox->setChecked(calculate);
  slotOptionsBoundingBox();
}

void Kpl::setPathOption(bool absolute)
{
  saveAbsPath->setChecked(absolute);
  slotOptionsSaveAbsPath();
}

void Kpl::setPageFormat(double w, double h)
{
  KplSettings::setFormatX(w);
  KplSettings::setFormatY(h);
  QSizeF size(w, h);
  if (size != doc->scene()->size())
    doc->scene()->setSize(size);
}

void Kpl::setColors(const QString& colFrame, const QString& colGrid,
                    const QString& colData)
{
  KplSettings::setColFrame(KplItem::rgb(colFrame));
  KplSettings::setColGrid(KplItem::rgb(colGrid));
  KplSettings::setColData(KplItem::rgb(colData));
}

void Kpl::setOutputFormat(const QString& pres, int prec)
{
  QString s("eEfgG");
  int i = s.indexOf(pres[0]);
  if (i >= 0)
    KplSettings::setPresentation(i);
  KplSettings::setPrecision(prec);
}

void Kpl::setAutoplot(double x0, double w, double y0, double h,
                      int gridmode, const QString& sx, const QString& sy,
                      const QString& sh, bool pathHeading, int ix, int iy,
                      int ie, bool logx, bool logy, bool errbars, int symb,
                      bool autoNorm, bool autoFit, double relSize)
{
  KplSettings::setX0(x0);
  KplSettings::setWidth(w);
  KplSettings::setY0(y0);
  KplSettings::setHeight(h);
  KplSettings::setGridMode(gridmode);
  KplSettings::setTextX(sx);
  KplSettings::setTextY(sy);
  KplSettings::setTextH(sh);
  KplSettings::setAutoPathHeader(pathHeading);
  KplSettings::setIX(ix);
  KplSettings::setIY(iy);
  KplSettings::setIE(ie);
  KplSettings::setLogX(logx);
  KplSettings::setLogY(logy);
  KplSettings::setErrBars(errbars);
  KplSettings::setSymbol(symb);
  KplSettings::setNormalization(autoNorm);
  KplSettings::setAutoFit(autoFit);
  KplSettings::setRelSize(qRound(100 * relSize));
}

void Kpl::setAutofit(const QString& path, const QString& name,
                     bool nonLin, bool errCol,
                     const QString& errModPath, const QString& errModName,
                     int itMax, double tol, int symb, const QString& color,
                     bool showDlg, bool savePar, bool follow)
{
  KplSettings::setFitPath(Utils::url(path));
  KplSettings::setFitName(name);
  KplSettings::setFitNonLinear(nonLin);
  if (nonLin) {
    KplSettings::setFitMaxIterations(itMax);
    KplSettings::setFitTolerance(tol);
  }
  KplSettings::setFitErrorColumn(errCol);
  if (!errCol) {
    KplSettings::setErrorModelPath(errModPath);
    KplSettings::setErrorModelName(errModName);
  }
  KplSettings::setFitSymbol(symb);
  KplSettings::setFitColor(KplItem::rgb(color));
  KplSettings::setFitShowDialog(showDlg);
  KplSettings::setFitSaveParam(savePar);
  KplSettings::setFitFollow(follow);
}

bool Kpl::loadAutofitPar(const QString& url)
{
  bool success = doc->loadPar(Utils::url(url), this, doc->pFit(),
                              KplUtils::decimalCharacter());
  if (success)
    KplSettings::setFitParam(FunItem::parToString(doc->pFit(),
                                                  doc->presentation(),
                                                  KplSettings::precision()));
  return success;
}

bool Kpl::setAutofitPar(int iPar, double value)
{
  if ((iPar < 0) || (iPar >= FunItem::NumParMax))
    return false;
  doc->pFit()[iPar] = value;
  KplSettings::setFitParam(FunItem::parToString(doc->pFit(),
                                                doc->presentation(),
                                                KplSettings::precision()));
  return true;
}

bool Kpl::enableFitPar(int iPar, bool fit)
{
  if ((iPar < 0) || (iPar >= FunItem::NumParMax))
    return false;
  doc->fitPar()[iPar] = fit;
  QStringList list;
  for (int i = 0; i < FunItem::NumParMax; ++i)
    list.append(doc->fitPar()[i] ? "1" : "0");
  KplSettings::setFitEnableParam(list.join(" "));
  return true;
}

bool Kpl::loadErrModPar(const QString& url, int iArr)
{
  bool success = false;
  if (iArr && (iArr >= arr.count()))
    return success;
  double p[FunItem::NumParMax];
  success = doc->loadPar(Utils::url(url), this, p,
                         KplUtils::decimalCharacter());
  if (success) {
    QString s = FunItem::parToString(p, doc->presentation(),
                                     KplSettings::precision());
    if (iArr < arr.count())
      KSharedConfig::openConfig()->group(QString("DataErrors%1").arg(iArr)).
        writeEntry("ErrorModelParam", s);
    if (!iArr)
      KplSettings::setErrorModelParam(s);
  }
  return success;
}

bool Kpl::setErrModPar(int iPar, double value, int iArr)
{
  if ((iPar < 0) || (iPar >= FunItem::NumParMax))
    return false;
  if (iArr && (iArr >= arr.count()))
    return false;
  QString groupName = QString("DataErrors%1").arg(iArr);
  double p[FunItem::NumParMax];
  FunItem::stringToPar(((iArr < arr.count()) &&
                        KSharedConfig::openConfig()->hasGroup(groupName)) ?
    KSharedConfig::openConfig()->group(groupName).readEntry("ErrorModelParam",
      KplSettings::errorModelParam()) : KplSettings::errorModelParam(), p);
  p[iPar] = value;
  QString s = FunItem::parToString(p, doc->presentation(),
                                   KplSettings::precision());
  if (iArr < arr.count())
    KSharedConfig::openConfig()->group(QString("DataErrors%1").arg(iArr)).
      writeEntry("ErrorModelParam", s);
  if (!iArr)
    KplSettings::setErrorModelParam(s);
  return true;
}

bool Kpl::deleteItem(int i)
{
  if ((i < 0) || (i >= itemCount))
    return false;
  doc->undoStack()->push(new DeleteCommand(doc->scene()->
    items(Qt::AscendingOrder).at(i), doc));
  return true;
}

bool Kpl::moveItem(int is, int id)
{
  if ((is < 0) || (id < 0) || (is == id) || (is >= itemCount) ||
      (id >= itemCount))
    return false;
  doc->undoStack()->push(new MoveCommand(is, id, doc));
  return true;
}

bool Kpl::activateItem(int i, bool active)
{
  if ((i < 0) || (i >= itemCount))
    return false;
  QGraphicsItem* item = doc->scene()->items(Qt::AscendingOrder).at(i);
  if (active == item->isVisible())
    return false;
  doc->undoStack()->push(new ToggleActiveCommand(item, doc));
  return true;
}

void Kpl::newFrameItem(double x0, double w, double y0, double h,
                       double xmin, double xmax, double ymin, double ymax,
                       bool logx, bool logy, int ndigx, int ndigy,
                       const QString& colFrame, const QString& colGrid,
                       double xtic, double ytic, int mticx, int mticy,
                       int gridmode, int iex, int iey,
                       const QString& sx, const QString& sy,
                       const QString& sh, double relSize, double xOff,
                       double yOff, bool xBottom, bool xTop, bool yLeft,
                       bool yRight, bool collapsed, bool visible)
{
  doc->insertItem(new FrameItem(x0, y0, w, h, KplItem::rgb(colFrame),
                                KplItem::rgb(colGrid), logx, logy,
                                FrameItem::GridMode(gridmode), xmax, ymax, sx,
                                sy, sh, relSize, ndigx, ndigy, mticx, mticy,
                                iex, iey, xmin, ymin, xtic, ytic, collapsed,
                                visible, xOff, yOff, xBottom, xTop, yLeft,
                                yRight));
}

void Kpl::newAutoFrameItem()
{
  doc->insertItem(new FrameItem(KplSettings::x0(), KplSettings::y0(),
                                KplSettings::width(), KplSettings::height(),
                                KplSettings::colFrame().rgb(),
                                KplSettings::colGrid().rgb(),
                                KplSettings::logX(), KplSettings::logY(),
                                FrameItem::GridMode(KplSettings::gridMode()),
                                KplSettings::width(), KplSettings::height(),
                                KplSettings::textX(), KplSettings::textY(),
                                KplSettings::textH(),
                                0.01 * KplSettings::relSize()));
}

bool Kpl::newArrayItem(const QString& path, int ix, int iy, int ie,
                       int istart, int n, bool errbars, int symbol,
                       const QString& color, double fx, double fy, int idec,
                       int fillStyle, double xOff, double yOff, double relSize,
                       bool visible)
{
  ArrayItem* item = new ArrayItem(KplItem::rgb(color), ix, iy, ie, errbars,
                                  KplItem::Symbol(symbol),
                                  static_cast<ArrayItem::DecimalCharacter>(
                                    idec), Utils::url(path), fx, fy, 0, n,
                                  istart, KplItem::FillStyle(fillStyle), xOff,
                                  yOff, visible, false, relSize);
  bool success = item->numRows();
  if (success)
    doc->insertItem(item);
  else
    delete item;
  return success;
}

bool Kpl::newFunItem(const QString& path, const QString& name, double xmin,
                     double xmax, double dx, int symbol, const QString& color,
                     double fx, double fy, int fillStyle, double xOff,
                     double yOff, double relSize, bool visible)
{
  FunItem* item = new FunItem(Utils::url(path), name, KplItem::rgb(color),
                              KplItem::Symbol(symbol), fx, fy, 0, xmin, xmax,
                              dx, visible, KplItem::FillStyle(fillStyle),
                              xOff, yOff, relSize);
  bool success = item->libY();
  if (success)
    doc->insertItem(item);
  else
    delete item;
  return success;
}

bool Kpl::newParFunItem(const QString& pathx, const QString& namex,
                        const QString& pathy, const QString& namey,
                        double tmin, double tmax, double dt, int symbol,
                        const QString& color, double fx, double fy,
                        int fillStyle, double xOff, double yOff,
                        double relSize, bool visible)
{
  ParFunItem* item = new ParFunItem(Utils::url(pathx), namex,
                                    Utils::url(pathy), namey,
                                    KplItem::rgb(color),
                                    KplItem::Symbol(symbol), fx, fy, 0,
                                    tmin, tmax, dt, visible,
                                    KplItem::FillStyle(fillStyle), xOff, yOff,
                                    relSize);
  bool success = (item->libY() && item->libY());
  if (success)
    doc->insertItem(item);
  else
    delete item;
  return success;
}

void Kpl::newSplineItem(int deriv, double low, double xmin, double xmax,
                        double dx, int symbol, const QString& color, double fx,
                        double fy, int k, int fillStyle, double xOff,
                        double yOff, double relSize, bool visible)
{
  doc->insertItem(new SplineItem(k, deriv, low, KplItem::rgb(color),
                                 KplItem::Symbol(symbol), fx, fy, 0, xmin,
                                 xmax, dx, visible,
                                 KplItem::FillStyle(fillStyle), xOff, yOff,
                                 relSize));
}

bool Kpl::newArray3DItem(double x0, double w, double y0, double h,
                         double xmin, double xmax, double ymin, double ymax,
                         double zmin, double zmax, bool logx, bool logy,
                         bool logz, int ndigx, int ndigy, int ndigz,
                         const QString& colFrame, const QString& colGrid,
                         const QString& colData, double xtic, double ytic,
                         double ztic, int mticx, int mticy, int mticz,
                         bool frame, int gridmode, int mode3D,
                         double xBox, double yBox, double zBox, double xRef,
                         double yRef, double phi, double theta, int iex,
                         int iey, int iez, const QString& sx,
                         const QString& sy, const QString& sz,
                         const QString& sh, double relSize,
                         const QString& path, int ix, int iy, int iz, int ie,
                         int istart, int n, bool errbars, double smf,
                         double dx, double dz, double fx, double fy,
                         double fz, int idec, bool collapsed, bool visible)
{
  Array3DItem* item = new Array3DItem(x0, y0, w, h, KplItem::rgb(colFrame),
                                      KplItem::rgb(colGrid),
                                      KplItem::rgb(colData), logx, logy, logz,
                                      FrameItem::GridMode(gridmode),
                                      ThreeDItem::Mode3D(mode3D), xmax, ymax,
                                      zmax, sx, sy, sz, sh, relSize, ix, iy,
                                      iz, ie, errbars,
                                      static_cast<ArrayItem::DecimalCharacter>(
                                        idec), ndigx,  ndigy, ndigz, mticx,
                                      mticy, mticz, iex, iey, iez, xmin, ymin,
                                      zmin, xtic, ytic, ztic, dx, dz, phi,
                                      theta, xBox, yBox, zBox, xRef, yRef, fx,
                                      fy, fz, istart, n, smf, Utils::url(path),
                                      frame, false, collapsed, visible);
  bool success = item->numRows();
  if (success)
    doc->insertItem(item);
  else
    delete item;
  return success;
}

bool Kpl::newFun3DItem(double x0, double w, double y0, double h,
                       double xmin, double xmax, double ymin, double ymax,
                       double zmin, double zmax, bool logx, bool logy,
                       bool logz, int ndigx, int ndigy, int ndigz,
                       const QString& colFrame, const QString& colGrid,
                       const QString& colData, double xtic, double ytic,
                       double ztic, int mticx, int mticy, int mticz,
                       bool frame, int gridmode, int mode3D, double xBox,
                       double yBox, double zBox, double xRef, double yRef,
                       double phi, double theta, int iex, int iey, int iez,
                       const QString& sx, const QString& sy,
                       const QString& sz, const QString& sh, double relSize,
                       const QString& path, const QString& name,
                       double dx, double dz, double fx, double fy,
                       double fz, bool collapsed, bool visible)
{
  Fun3DItem* item = new Fun3DItem(x0, y0, w, h, KplItem::rgb(colFrame),
                                  KplItem::rgb(colGrid), KplItem::rgb(colData),
                                  logx, logy, logz,
                                  FrameItem::GridMode(gridmode),
                                  ThreeDItem::Mode3D(mode3D), xmax, ymax, zmax,
                                  sx, sy, sz, sh, relSize, ndigx, ndigy, ndigz,
                                  mticx, mticy, mticz, iex, iey, iez, xmin,
                                  ymin, zmin, xtic, ytic, ztic, dx, dz, phi,
                                  theta, xBox, yBox, zBox, xRef, yRef, fx, fy,
                                  fz, name, Utils::url(path), frame,
                                  collapsed, visible);
  bool success = item->libY();
  if (success)
    doc->insertItem(item);
  else
    delete item;
  return success;
}

void Kpl::newImageItem(const QString& path, double x, double y, double fx,
                       double fy, double dir, bool visible)
{
  doc->insertItem(new ImageItem(x, y, 0, Utils::url(path), visible, fx, fy,
                                dir));
}

void Kpl::newLegendItem(int symbol, const QString& colSymb, double x, double y,
                        const QString& text, const QString& colText,
                        double xoff, double yoff, double relSize,
                        double symbolSize, bool visible)
{
  doc->insertItem(new LegendItem(x, y, KplItem::rgb(colSymb),
             KplItem::rgb(colText), KplItem::Symbol(symbol), xoff, yoff, 0,
             text, visible, relSize, symbolSize));
}

void Kpl::newTextItem(const QString& text, const QString& color, int alignment,
                      double x, double y, double dir, double relSize,
                      bool visible)
{
  doc->insertItem(new TextItem(x, y, KplItem::rgb(color), 0, text, visible,
                               TextItem::Alignment(alignment), relSize, dir));
}

void Kpl::newLineItem(int symbol, const QString& color, double x, double y,
                      double dir, double len, double relSize, bool visible)
{
  doc->insertItem(new LineItem(x, y, KplItem::rgb(color), 0, len, visible,
                               KplItem::Symbol(symbol), relSize, dir));
}

void Kpl::newArrowItem(int symbol, const QString& color, double x, double y,
                       double dir, double len, double relSize,
                       bool doubleHeaded, bool visible)
{
  doc->insertItem(new ArrowItem(x, y, KplItem::rgb(color), 0, len, visible,
                                doubleHeaded, KplItem::Symbol(symbol), relSize,
                                dir));
}

void Kpl::newArcItem(int symbol, const QString& color, double x, double y,
                     double w, double h, double start, double span,
                     double relSize, double dir, bool visible)
{
  doc->insertItem(new ArcItem(x, y, KplItem::rgb(color), 0, w, h, visible,
                              start, span, KplItem::Symbol(symbol), relSize,
                              dir));
}

void Kpl::newRectangleItem(int symbol, const QString& color, double x, double y,
                           double w, double h, double relSize, int fillStyle,
                           double dir, bool visible)
{
  doc->insertItem(new RectItem(x, y, KplItem::rgb(color), 0, w, h, visible,
                               KplItem::FillStyle(fillStyle),
                               KplItem::Symbol(symbol), relSize, dir));
}

void Kpl::newEllipseItem(int symbol, const QString& color, double x, double y,
                         double w, double h, double relSize, int fillStyle,
                         double dir, bool visible)
{
  doc->insertItem(new EllipseItem(x, y, KplItem::rgb(color), 0, w, h, visible,
                                  KplItem::FillStyle(fillStyle),
                                  KplItem::Symbol(symbol), relSize, dir));
}

void Kpl::newScaleBarItem(const QString& text, const QString& color,
                          double x, double y, double len, double relSize,
                          bool vertical, bool visible)
{
  doc->insertItem(new ScaleBarItem(x, y, KplItem::rgb(color), 0, len, text,
                                   visible, vertical, relSize));
}

bool Kpl::autoScaleFrame(int i, bool autoNorm)
{
  if ((i < 0) || (i >= itemCount))
    return false;
  QGraphicsItem* o = doc->scene()->items(Qt::AscendingOrder).at(i);
  if (o->type() != FrameItem::Type)
    return false;
  FrameItem* fi = new FrameItem(*qgraphicsitem_cast<FrameItem*>(o));
  doc->swapParents(fi, o);
  fi->autoScale(autoNorm);
  doc->swapParents(o, fi);
  doc->undoStack()->push(new EditCommand(fi, o, doc));
  return true;
}

bool Kpl::loadPar(const QString& url, int i, bool yFun)
{
  if ((i < 0) || (i >= itemCount))
    return false;
  QGraphicsItem* o = doc->scene()->items(Qt::AscendingOrder).at(i);
  QGraphicsItem* n = 0;
  switch (o->type()) {
    case FunItem::Type:
      if (yFun) {
        FunItem* fi = new FunItem(*qgraphicsitem_cast<FunItem*>(o));
        if (doc->loadPar(Utils::url(url), this, fi->parY(),
                         KplUtils::decimalCharacter()))
          n = fi;
        else
          delete fi;
      }
      break;
    case ParFunItem::Type: {
      ParFunItem* fi = new ParFunItem(*qgraphicsitem_cast<ParFunItem*>(o));
      if (doc->loadPar(Utils::url(url), this,
          yFun ? fi->parY() : fi->parX(), KplUtils::decimalCharacter()))
        n = fi;
      else
        delete fi;
      break;
    }
    case Fun3DItem::Type:
      if (yFun) {
        Fun3DItem* fi = new Fun3DItem(*qgraphicsitem_cast<Fun3DItem*>(o));
        if (doc->loadPar(Utils::url(url), this, fi->parY(),
                         KplUtils::decimalCharacter()))
          n = fi;
        else
          delete fi;
      }
  }
  if (n)
    doc->undoStack()->push(new EditCommand(n, o, doc));
  return n;
}

bool Kpl::setPar(int i, int iPar, double value, bool yFun)
{
  if ((i < 0) || (i >= itemCount) || (iPar < 0) ||
      (iPar >= FunItem::NumParMax))
    return false;
  QGraphicsItem* o = doc->scene()->items(Qt::AscendingOrder).at(i);
  QGraphicsItem* n = 0;
  switch (o->type()) {
    case FunItem::Type:
      if (yFun) {
        FunItem* fi = new FunItem(*qgraphicsitem_cast<FunItem*>(o));
        fi->setPar(iPar, value, yFun);
        n = fi;
      }
      break;
    case ParFunItem::Type: {
      ParFunItem* fi = new ParFunItem(*qgraphicsitem_cast<ParFunItem*>(o));
      fi->setPar(iPar, value, yFun);
      n = fi;
      break;
    }
    case Fun3DItem::Type: {
      Fun3DItem* fi = new Fun3DItem(*qgraphicsitem_cast<Fun3DItem*>(o));
      fi->setPar(iPar, value);
      n = fi;
    }
  }
  if (n)
    doc->undoStack()->push(new EditCommand(n, o, doc));
  return n;
}

bool Kpl::setArrayRange(int i, int iStart, int np)
{
  if ((iStart < 0) || (np < 1) || ( i < 0) || (i >= itemCount))
    return false;
  QGraphicsItem* o = doc->scene()->items(Qt::AscendingOrder).at(i);
  QGraphicsItem* n = 0;
  switch (o->type()) {
    case ArrayItem::Type: {
      ArrayItem* ai0 = qgraphicsitem_cast<ArrayItem*>(o);
      if ((iStart > (ai0->numRows() - 1)) || ((iStart + np) > ai0->numRows()))
        return false;
      ArrayItem* ai = new ArrayItem(*ai0);
      ai->setRowStart(iStart);
      ai->setCount(np);
      n = ai;
      break;
    }
    case Array3DItem::Type:
      Array3DItem* ai0 = qgraphicsitem_cast<Array3DItem*>(o);
      if ((iStart > (ai0->numRows() - 1)) || ((iStart + np) > ai0->numRows()))
        return false;
      Array3DItem* ai = new Array3DItem(*ai0);
      ai->setRowStart(iStart);
      ai->setCount(np);
      n = ai;
  }
  if (n)
    doc->undoStack()->push(new EditCommand(n, o, doc));
  return n;
}

bool Kpl::addFitItems(bool erase, int iArr, int iFun, bool errCol,
                      const QString& errModPath, const QString& errModName)
{
  if ((iArr < 0) || (iArr >= itemCount) || (iFun < 0) || (iFun >= itemCount))
    return false;
  QList<QGraphicsItem*> list = doc->scene()->items(Qt::AscendingOrder);
  if ((list.at(iArr)->type() != ArrayItem::Type) ||
      (list.at(iFun)->type() != FunItem::Type))
    return false;
  if (erase) {
    arr.clear();
    fun.clear();
  }
  KConfigGroup config = KSharedConfig::openConfig()->
      group(QString("DataErrors%1").arg(arr.count()));
  config.writeEntry("FitErrorColumn", bool(errCol));
  if (!errCol) {
    config.writeEntry("ErrorModelPath", errModPath);
    config.writeEntry("ErrorModelName", errModName);
  }
  arr.append(qgraphicsitem_cast<ArrayItem*>(list.at(iArr)));
  fun.append(qgraphicsitem_cast<FunItem*>(list.at(iFun)));
  return true;
}

bool Kpl::addFunFitItem(int iFun)
{
  if ((iFun < 0) || (iFun >= itemCount))
    return false;
  QList<QGraphicsItem*> list = doc->scene()->items(Qt::AscendingOrder);
  if (list.at(iFun)->type() != FunItem::Type)
    return false;
  fun.append(qgraphicsitem_cast<FunItem*>(list.at(iFun)));
  return true;
}

bool Kpl::fit(bool nonLin, bool savePar, bool follow)
{
  if (!arr.count())
    return false;
  doc->fitNonLinear() = nonLin;
  int mode = 0;
  if (savePar)
    mode |= FitDlg::SavePar;
  if (follow)
    mode |= FitDlg::Follow;
  doc->chiSquare() = 0;
  FitDlg dlg(this, doc, &arr, fun, mode);
  dlg.slotFit();
  dlg.apply();
  return true;
}

bool Kpl::addSplineFitItem(bool erase, int iSpl)
{
  QList<QGraphicsItem*> list = doc->scene()->items(Qt::AscendingOrder);
  if ((iSpl < 0) || (iSpl >= itemCount) ||
      (list.at(iSpl)->type() != SplineItem::Type))
    return false;
  if (erase)
    spl.clear();
  spl.append(qgraphicsitem_cast<SplineItem*>(list.at(iSpl)));
  return true;
}

bool Kpl::splineFit(int iArr, int deg, double smf, double xmin,
                    double xmax, bool errCol, const QString& errModPath,
                    const QString& errModName)
{
  QList<QGraphicsItem*> list = doc->scene()->items(Qt::AscendingOrder);
  if (!spl.count() || (iArr < 0) || (iArr >= itemCount) ||
      (list.at(iArr)->type() != ArrayItem::Type))
    return false;
  ArrayItem* ai = qgraphicsitem_cast<ArrayItem*>(list.at(iArr));
  if (errCol && (ai->colErr() >= ai->numCols()))
    return false;
  doc->chiSquare() = 0;
  SplFitDlg dlg(this, doc, ai, spl, 0);
  FunItem::DataErrorStruct err = doc->errorModel();
  err.fitErrCol = errCol;
  err.errModPath = errModPath;
  err.errModName = errModName;
  if (dlg.fit(deg, smf, xmin, xmax, &err) < 4) {
    dlg.apply();
    return true;
  } else
    return false;
}

double Kpl::chiSquare()
{
  return doc->chiSquare();
}

bool Kpl::saveFitPar(const QString& url)
{
  return doc->saveParameter(this, Utils::url(url), doc->pLast(),
                            doc->pErr());
}

bool Kpl::exportValues(const QString& url, int i)
{
  if ((i < 0) || (i >= itemCount))
    return false;
  QGraphicsItem* item = doc->scene()->items(Qt::AscendingOrder).at(i);
  int ityp = item->type();
  if ((ityp != FunItem::Type) && (ityp != ParFunItem::Type) &&
      (ityp != SplineItem::Type) && (ityp != KplItem::Function3D) &&
      (ityp != Array3DItem::Type))
    return false;
  QUrl u(url);
  QFile f(u.isLocalFile() ? u.path() : Utils::tempFileName());
  if (f.open(QIODevice::WriteOnly)) {
    QTextStream ts(&f);
    char sep = KplDoc::separator();
    char format = KplDoc::presentation();
    int prec = KplSettings::precision();
    switch (ityp) {
      case FunItem::Type:
        qgraphicsitem_cast<FunItem*>(item)->exportTable(ts, sep, format, prec);
        break;
      case ParFunItem::Type:
        qgraphicsitem_cast<ParFunItem*>(item)->exportTable(ts, sep, format,
                                                           prec);
        break;
      case SplineItem::Type:
        qgraphicsitem_cast<SplineItem*>(item)->exportTable(ts, sep, format,
                                                           prec);
        break;
      case Fun3DItem::Type:
        qgraphicsitem_cast<Fun3DItem*>(item)->exportTable(ts, sep, format,
                                                          prec);
        break;
      case Array3DItem::Type:
        qgraphicsitem_cast<Array3DItem*>(item)->exportTable(ts, sep, format,
                                                            prec);
    }
    f.close();
    if (!u.isLocalFile())
      Utils::move(f.fileName(), u, this);
    return true;
  } else {
    KMessageBox::sorry(this, i18n("Writing to this file not possible."));
    return false;
  }
}

bool Kpl::exportImageUrl(const QUrl& url)
{
  if (url.isEmpty())
    return false;
  view->viewport()->setCursor(Qt::WaitCursor);
  bool success = doc->scene()->exportImage(url, KplSettings::boundingBox());
  if (success)
    KplUtils::setCurrentDir(url);
  view->viewport()->setCursor(Qt::CrossCursor);
  return success;
}

void Kpl::generatePS(bool landscape)
{
  statusMsg(i18n("Generating PostScript file..."));
  QUrl u;
  if (KplUtils::getWriteUrl(this, u, KplSettings::currentDir(),
                            "*.eps Encapsulated PostScript (*.eps);;"
                            "*.ps PostScript (*.ps);;* (*)"))
    if (!savePSUrl(u, landscape))
      KMessageBox::sorry(this, i18n("Writing to this file not possible."));
  ready();
}

void Kpl::initStatusBar()
{
  statusAddLabel = new QLabel();
  QFontMetrics fm = statusBar()->fontMetrics();
  int wM = fm.boundingRect("M").width();
  statusAddLabel->
    setFixedWidth(qMax(fm.boundingRect(i18n("Add")).width(),
                       fm.boundingRect(i18n("Replace")).width()) + wM);
  statusBar()->addPermanentWidget(statusAddLabel);
  statusItemWidth = qMin(width() / 3, 50 * wM);
  statusItemLabel->setFixedWidth(statusItemWidth + 3);
  statusBar()->addPermanentWidget(statusItemLabel);
  statusPosLabel->setFixedWidth(fm.boundingRect(
    " 10.00 cm,  15.00 cm ").width() + 3);
  statusBar()->addPermanentWidget(statusPosLabel);
  statusValWidth = fm.boundingRect(
    " x = -1.234e-123, y = -1.234e-123 ").width() + 3;
  statusBar()->addPermanentWidget(statusValLabel);
  statusMsgLabel->setText(i18n("Ready."));
  statusBar()->addWidget(statusMsgLabel);
}

void Kpl::moveItem(int id)
{
  doc->undoStack()->push(new MoveCommand(sourceIndex, id, doc));
  itemList->clearSelection();
  QItemSelectionModel* sel = itemList->selectionModel();
  QList<QGraphicsItem*> list = doc->scene()->items(Qt::AscendingOrder);
  for (int r = 0; r < doc->rowCount(); ++r) {
    QModelIndex p = doc->index(r, 0);
    if (p.isValid()) {
      if (qRound(static_cast<QGraphicsItem*>(p.internalPointer())->
          zValue()) == id) {
        sel->select(p,
                    QItemSelectionModel::Select | QItemSelectionModel::Rows);
        break;
      }
      for (int r2 = 0; r2 < doc->rowCount(p); ++r2) {
        QModelIndex i = doc->index(r2, 0, p);
        if (i.isValid())
          if (qRound(static_cast<QGraphicsItem*>(i.internalPointer())->
              zValue()) == id) {
            sel->select(i, QItemSelectionModel::Select |
                           QItemSelectionModel::Rows);
            break;
          }
      }
    }
  }
}

void Kpl::openRead(int type)
{
  statusMsg(i18n("Opening file..."));
  ArrayItem::DecimalCharacter idec = KplUtils::decimalCharacter();
  QList<QUrl> list = (type == KplDoc::Data ?
                      FileDialog::getReadUrls(idec, this) :
                      FileDialog::getReadUrls(this));
  if (!list.isEmpty())
    openList(&list, idec, type);
  ready();
}

void Kpl::preview(bool landscape)
{
  statusMsg(i18n("Generating PostScript preview..."));
  QUrl u;
  savePSUrl(u, landscape);
  ready();
}

bool Kpl::queryClose()
{
  if ((!doc->undoStack()->isClean()) &&
      doc->scene()->items(Qt::AscendingOrder).count()) {
    do {
      int i = KMessageBox::warningYesNoCancel(this,
        i18n("The current plot has been changed.\nSave file?"));
      if (i == KMessageBox::Cancel)
        return false;
      if (i == KMessageBox::No)
        break;
    } while (!save(true));
    doc->undoStack()->clear();
  }
  recentFiles->saveEntries(KSharedConfig::openConfig()->group("Recent Files"));
  KplSettings::setZoom(view->zoomFactor());
  KplSettings::self()->save();
  return true;
}

void Kpl::readProperties(const KConfigGroup& cfg)
{
  QString url = cfg.readPathEntry("lastUrl", "");
  if (!url.isEmpty())
    if (doc->openDocument(Utils::url(url), KplUtils::decimalCharacter(),
        KplDoc::Unknown))
      updWin();
}

void Kpl::ready()
{
  statusMsg(i18n("Ready."));
}

bool Kpl::save(bool current)
{
  bool success = false;
  if ((doc->url().fileName() != i18n("Untitled")) && current &&
      (doc->fileType() == KplDoc::Plot)) {
    success = doc->saveDocument(doc->url(), KplSettings::saveAbsPath());
    if (success)
      doc->setFileType(KplDoc::Plot);
  } else {
    QUrl url;
    if (KplUtils::getWriteUrl(this, url, KplSettings::currentDir(),
                              "*.plo (*.plo);;* (*)"))
      success = saveAs(url);
  }
  return success;
}

bool Kpl::saveAs(const QUrl& url)
{
  bool success = doc->saveDocument(url, KplSettings::saveAbsPath());
  if (success) {
    doc->setUrl(url);
    doc->setFileType(KplDoc::Plot);
    updWin();
  }
  return success;
}

void Kpl::saveProperties(KConfigGroup& cfg)
{
  if (doc->url().fileName() != i18n("Untitled"))
    cfg.writePathEntry("lastUrl", doc->url().toDisplayString());
}

bool Kpl::savePSUrl(const QUrl& url, bool landscape)
{
  view->viewport()->setCursor(Qt::WaitCursor);
  bool success = doc->scene()->savePS(url, landscape,
                                      KplSettings::boundingBox());
  if (success && (!url.isEmpty()))
    KplUtils::setCurrentDir(url);
  view->viewport()->setCursor(Qt::CrossCursor);
  return success;
}

void Kpl::clearStatus()
{
  if (statusBar()->isVisible()) {
    statusValLabel->setText("");
    statusPosLabel->setText("");
    statusItemLabel->setText("");
  }
}

void Kpl::addPreview(const QString& filename)
{
  previews << filename;
}

void Kpl::select()
{
  itemList->clearSelection();
  QItemSelectionModel* sel = itemList->selectionModel();
  QList<QGraphicsItem*> list = doc->scene()->selectedItems();
  for (int r = 0; r < doc->rowCount(); ++r) {
    QModelIndex p = doc->index(r, 0);
    if (p.isValid()) {
      foreach (QGraphicsItem* it, list)
        if (it == p.internalPointer()) {
          sel->select(p, QItemSelectionModel::Select |
                      QItemSelectionModel::Rows);
          break;
        }
      for (int r2 = 0; r2 < doc->rowCount(p); ++r2) {
        QModelIndex i = doc->index(r2, 0, p);
        if (i.isValid())
          foreach (QGraphicsItem* it, list)
            if (it == i.internalPointer()) {
              sel->select(i, QItemSelectionModel::Select |
                          QItemSelectionModel::Rows);
              break;
            }
      }
    }
  }
}

void Kpl::settingsChanged()
{
  QSizeF size(KplSettings::formatX(), KplSettings::formatY());
  if (size != doc->scene()->size())
    doc->scene()->setSize(size);
  showSource->setChecked(KplSettings::showSource());
  bool b = KplSettings::addItems();
  addItems->setChecked(b);
  statusAddLabel->setText(i18n(b ? "Add" : "Replace"));
  boundingBox->setChecked(KplSettings::boundingBox());
  saveAbsPath->setChecked(KplSettings::saveAbsPath());
  warning->setChecked(KplSettings::warning());
  ThreadWeaver::Queue::instance()->
    setMaximumNumberOfThreads(KplSettings::multithreading() ?
                              qMax(1, QThread::idealThreadCount()) : 1);
}

void Kpl::slotActive()
{
  if (currentIndex.isValid())
    doc->undoStack()->push(new ToggleActiveCommand(static_cast<QGraphicsItem*>(
      currentIndex.internalPointer()), doc));
}

void Kpl::slotCopy()
{
  doc->copyItems(itemList->selectionModel()->selectedRows());
}

void Kpl::slotCut()
{
  doc->copyItems(itemList->selectionModel()->selectedRows(), true);
}

void Kpl::slotDelete()
{
  doc->removeItems(itemList->selectionModel()->selectedRows());
}

void Kpl::slotEdit()
{
  statusMsg(i18n("Editing item..."));
  doc->editItem(itemList->selectionModel()->selectedRows(),
                pos.isNull() ? itemList->mapToGlobal(itemList->pos()) : pos);
  ready();
}

void Kpl::slotEditBottom()
{
  moveItem(itemCount - 1);
}

void Kpl::slotEditDown()
{
  moveItem(sourceIndex + 1);
}

void Kpl::slotEditItems()
{
  itemWidget->show();
}

void Kpl::slotEditTop()
{
  moveItem(0);
}

void Kpl::slotEditUp()
{
  moveItem(sourceIndex - 1);
}

void Kpl::slotFileNew()
{
  if (queryClose())
    doc->newDocument();
  ready();
}

void Kpl::slotFileOpenData()
{
  openRead(KplDoc::Data);
}

void Kpl::slotFileOpenPlot()
{
  openRead(KplDoc::Plot);
}

void Kpl::slotFileExport()
{
  statusMsg(i18n("Exporting to image file..."));
  QUrl u;
  if (KplUtils::getWriteUrl(this, u, KplSettings::currentDir(),
                            "*.pdf PDF (*.pdf);;*.png PNG (*.png);;"
                            "*.ppm PPM (*.ppm);;*.bmp BMP (*.bmp);;"
                            "*.jpg *.jpeg JPEG (*.jpg *.jpeg);;"
                            "*.tif *.tiff TIFF (*.tif *.tiff);;"
                            "*.svg SVG (*.svg);;* (*)"))
    if (!exportImageUrl(u))
      KMessageBox::sorry(this, i18n("Writing to this file not possible."));
  ready();
}

void Kpl::slotFilePreviewP()
{
  preview(false);
}

void Kpl::slotFilePreviewL()
{
  preview(true);
}

void Kpl::slotFilePrint()
{
  statusMsg(i18n("Printing..."));
  QPrinter p(QPrinter::HighResolution);
  p.setColorMode(QPrinter::Color);
  if (QPrintDialog(&p).exec() == QDialog::Accepted)
    doc->scene()->print(&p);
  ready();
}

void Kpl::slotFilePSP()
{
  generatePS(false);
}

void Kpl::slotFilePSL()
{
  generatePS(true);
}

void Kpl::slotFileSave()
{
  statusMsg(i18n("Saving file..."));
  save(true);
  ready();
}

void Kpl::slotFileSaveAs()
{
  statusMsg(i18n("Saving file under different name..."));
  save(false);
  ready();
}

void Kpl::slotFileNewWindow()
{
  (new Kpl)->show();
  ready();
}

void Kpl::slotFileQuit()
{
  statusMsg(i18n("Exiting..."));
  foreach (KMainWindow* w, memberList())
    if (!w->close()) {
      statusMsg(i18n("Ready."));
      break;
    }
}

void Kpl::slotFit()
{
  doc->fitItems(itemList->selectionModel()->selectedRows());
  ready();
}

void Kpl::slotNewItem()
{
  statusMsg(i18n("Inserting new item..."));
  int type;
  QPoint p = pos.isNull() ? itemList->mapToGlobal(itemList->pos()) : pos;
  NewItemDlg dlg(this, &type, p);
  if (dlg.exec()) {
    statusMsg(i18n("Editing new item..."));
    QModelIndexList indexes = itemList->selectionModel()->selectedRows();
    doc->newItem(type, indexes.isEmpty() ? -1 :
      static_cast<QGraphicsItem*>(indexes.first().internalPointer())->zValue(),
      p);
  }
  ready();
}

void Kpl::slotOpenRecent(const QUrl& url)
{
  statusMsg(i18n("Opening file..."));
  if (KplSettings::addItems() || (KplSettings::warning() ? queryClose() :
                                                           true))
    if (!doc->openDocument(url, KplUtils::decimalCharacter(), KplDoc::Unknown))
      recentFiles->removeUrl(url);
  ready();
}

void Kpl::slotOptionsAddData()
{
  bool b = addItems->isChecked();
  KplSettings::setAddItems(b);
  statusAddLabel->setText(i18n(b ? "Add" : "Replace"));
}

void Kpl::slotOptionsNotifications()
{
  statusMsg(i18n("Notification configuration..."));
  KNotifyConfigWidget::configure(this);
  ready();
}

void Kpl::slotOptionsPreferences()
{
  if (!KConfigDialog::showDialog("settings")) {
    SettingsDlg* dlg = new SettingsDlg(this, "settings",
                                       KplSettings::self(), doc);
    dlg->setAttribute(Qt::WA_DeleteOnClose);
    connect(dlg, SIGNAL(settingsChanged(const QString&)),
            SLOT(settingsChanged()));
    dlg->show();
  }
}

void Kpl::slotOptionsBoundingBox()
{
  KplSettings::setBoundingBox(boundingBox->isChecked());
}

void Kpl::slotOptionsSaveAbsPath()
{
  KplSettings::setSaveAbsPath(saveAbsPath->isChecked());
}

void Kpl::slotOptionsShowSource()
{
  KplSettings::setShowSource(showSource->isChecked());
}

void Kpl::slotOptionsWarning()
{
  KplSettings::setWarning(warning->isChecked());
}

void Kpl::slotPaste()
{
  const QMimeData* md = QApplication::clipboard()->mimeData();
  if (md->hasUrls()) {
    QModelIndexList indexes = itemList->selectionModel()->selectedRows();
    doc->insertItems(md->urls().first(), indexes.isEmpty() ? -1 :
      static_cast<QGraphicsItem*>(indexes.first().internalPointer())->
        zValue());
  }
}

void Kpl::slotPopup(const QPoint& p)
{
  pos = itemList->mapToGlobal(p);
  currentIndex = itemList->indexAt(p);
  if (currentIndex.isValid()) {
    itemList->clearSelection();
    itemList->setCurrentIndex(currentIndex);
    editActive->setChecked(static_cast<QGraphicsItem*>(currentIndex.
      internalPointer())->isVisible());
    static_cast<QMenu*>(factory()->container("popup_item", this))->exec(pos);
  } else
    static_cast<QMenu*>(factory()->container("popup", this))->exec(pos);
  pos = QPoint();
}

void Kpl::slotPopupItem(const QPoint& p)
{
  pos = view->mapToGlobal(p);
  QGraphicsItem* item = view->itemAt(p);
  if (item) {
    itemList->clearSelection();
    doc->scene()->clearSelection();
    item->setSelected(true);
    currentIndex = itemList->selectionModel()->selectedRows().first();
    editActive->setChecked(item->isVisible());
    static_cast<QMenu*>(factory()->container("popup_item", this))->exec(pos);
  } else
    static_cast<QMenu*>(factory()->container("popup", this))->exec(pos);
  pos = QPoint();
}

void Kpl::slotViewAutoReload()
{
  statusMsg(i18n("Reloading file periodically..."));
  RefreshDlg(this, timer).exec();
  ready();
}

void Kpl::slotViewReload()
{
  if (!reloadActive) {
    reloadActive = true;
    statusMsg(i18n("Reloading file..."));
    bool sav = KplSettings::addItems();
    KplSettings::setAddItems(false);
    doc->openDocument(QUrl(doc->url()), KplUtils::decimalCharacter(),
                      doc->fileType());
    KplSettings::setAddItems(sav);
    updWin();
    ready();
    reloadActive = false;
  }
}

void Kpl::slotViewZoom()
{
  statusMsg(i18n("Setting magnification..."));
  bool ok;
  int i = QInputDialog::getInt(this, i18n("Zoom"), i18n("Factor [%]"),
                               qRound(100 * view->zoomFactor()), 1, 2000, 1,
                               &ok);
  if (ok)
    view->setZoomFactor(0.01 * i);
  ready();
}

void Kpl::slotViewZoomIn()
{
  view->setZoomFactor(0.01 * (qMin(qRound(100 * view->zoomFactor()) + 20,
                                   2000)));
  ready();
}

void Kpl::slotViewZoomOut()
{
  view->setZoomFactor(0.01 * (qMax(qRound(100 * view->zoomFactor()) - 20, 1)));
  ready();
}

void Kpl::slotWatchFile()
{
  if (doc->url().isLocalFile() && 
      (doc->url().fileName() != i18n("Untitled")) &&
      (QFileInfo(doc->url().path()).lastModified() != doc->urlTime()))
    slotViewReload();
}

void Kpl::statusMsg(const QString& text)
{
  statusMsgLabel->setText(text);
}

void Kpl::updatePasteAction()
{
  bool plo = false;
  const QMimeData* md = QApplication::clipboard()->mimeData();
  if (md->hasUrls()) {
    QFileInfo fi(md->urls().first().path());
    plo = fi.suffix().toLower() == "plo";
  }
  editPaste->setEnabled(plo);
}

void Kpl::updStatus(const QPointF& p, const FrameItem* frame)
{
  if (statusBar()->isVisible()) {
    QPointF p2 = doc->scene()->mapInv(p);
    statusPosLabel->setText(QString(" %1 cm, %2 cm ").arg(p2.x(), 0, 'f', 2).
      arg(p2.y(), 0, 'f', 2));
    QFontMetrics fm = statusBar()->fontMetrics();
    if (frame) {
      p2 = frame->mapToUser(p);
      QPointF d = frame->mapToUser(p + QPoint(1, 1)) - p2;
      int xp = precision(p2.x(), d.x());
      int yp = precision(p2.y(), d.y());
      if ((xp != xPrec) || (yp != yPrec)) {
        xPrec = xp;
        yPrec = yp;
        static const double v = -1.234567890123456e-123;
        int w = fm.boundingRect(QString(" x = %1, y = %2 ").
                                arg(v, 0, 'g', xPrec).
                                arg(v, 0, 'g', yPrec)).width() + 3;
        if (w > statusValWidth) {
          statusValWidth = w;
          statusValLabel->setFixedWidth(statusValWidth);
        }
      }
      statusValLabel->setText(QString(" x = %1, y = %2 ").
        arg(p2.x() / frame->normX(), 0, 'g', xPrec).
        arg(p2.y() / frame->normY(), 0, 'g', yPrec));
    } else
      statusValLabel->setText("");
    QGraphicsItem* item = doc->scene()->itemAt(p, QTransform());
    QString s = "";
    if (item) {
      QString s2 = s = QString(" z = %1, %2   %3").arg(item->zValue()).
        arg(KplDoc::typeName(item)).arg(KplDoc::settings(item));
      if (item->type() == ArrayItem::Type) {
        ArrayItem* a = qgraphicsitem_cast<ArrayItem*>(item);
        if (a->count() <= 3001) {
          double* ax = a->data()[a->colX()];
          double* ay = a->data()[a->colY()];
          QList<QPolygonF> list = item->shape().toSubpathPolygons();
          for (int i = 0; i < list.count(); ++i) {
            if (list[i].containsPoint(p, Qt::OddEvenFill)) {
              int j = -1;
              int k = a->rowStart();
              do {
                double x = ax[++j + k];
                double y = ay[j + k];
                if (!(frame ? frame->valid(x, y) :
                              (FrameItem::valid(x) && FrameItem::valid(y)))) {
                  k++;
                  j--;
                }
              } while (j < i);
              s2 = s += QString(", i = %1").arg(j + k);
              break;
            }
          }
        }
      }
      s2 = s += " ";
      while (fm.boundingRect(s).width() > statusItemWidth) {
        s2.chop(1);
        s = s2 + "...";
      }
    }
    if (s != statusItemLabel->text())
      statusItemLabel->setText(s);
  }
}

void Kpl::updWin()
{
  QString filename = doc->url().fileName();
  bool untitled = (filename == i18n("Untitled"));
  viewReload->setEnabled(!untitled);
  if (!untitled) {
    recentFiles->addUrl(doc->url());
    filename = doc->url().isLocalFile() ? doc->url().path() : doc->url().url();
  }
  bool notClean = !doc->undoStack()->isClean();
  setCaption(filename, notClean);
  fileSave->setEnabled(notClean);
  itemCount = doc->scene()->items(Qt::AscendingOrder).count();
  fileSaveAs->setEnabled(itemCount);
  filePreviewP->setEnabled(itemCount);
  filePreviewL->setEnabled(itemCount);
  filePrint->setEnabled(itemCount);
  filePSP->setEnabled(itemCount);
  filePSL->setEnabled(itemCount);
  fileExport->setEnabled(itemCount);
  int nSel = 0;
  if (itemCount) {
    QModelIndexList list = itemList->selectionModel()->selectedRows();
    nSel = list.count();
    editFit->setEnabled(doc->fittable(list));
  } else
    editFit->setEnabled(false);
  editItem->setEnabled(nSel);
  editDelete->setEnabled(nSel);
  editCopy->setEnabled(nSel);
  editCut->setEnabled(nSel);
  bool single = nSel == 1;
  editActive->setEnabled(single);
  if (single) {
    currentIndex = itemList->selectionModel()->selectedRows().first();
    QGraphicsItem* item =
      static_cast<QGraphicsItem*>(currentIndex.internalPointer());
    editActive->setChecked(item->isVisible());
    sourceIndex = qRound(item->zValue());
  }
  bool up = single && sourceIndex;
  editUp->setEnabled(up);
  editTop->setEnabled(up);
  bool down = single && (sourceIndex < (itemCount - 1));
  editDown->setEnabled(down);
  editBottom->setEnabled(down);
}
