/***************************************************************************
                          splinedlg.cpp  -  description
                             -------------------
    begin                : Sun Mar 3 2002
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QComboBox>
#include <QDesktopServices>
#include <QDialogButtonBox>
#include <QFile>
#include <QGroupBox>
#include <QKeyEvent>
#include <QLabel>
#include <QLayout>
#include <QTextStream>
#include <KColorButton>
#include <KLocalizedString>
#include <KMessageBox>
#include "brushdlg.h"
#include "doubleedit.h"
#include "doublespinbox.h"
#include "kpl.h"
#include "kpldoc.h"
#include "kplsettings.h"
#include "kplutils.h"
#include "rootdlg.h"
#include "spinbox.h"
#include "splinedlg.h"
#include "splineitem.h"
#include "symboldlg.h"
#include "utils.h"

SplineDlg::SplineDlg(QWidget* parent, KplDoc* model, SplineItem* si0,
                     const QPoint& pos) :
 QDialog(parent), m(model), si(si0)
{
  sit = new SplineItem(*si);
  setWindowTitle(i18n("Smoothing spline"));
  QVBoxLayout* vbox = new QVBoxLayout(this);
  QGroupBox* g = new QGroupBox(i18n("Output"));
  QHBoxLayout* hbox = new QHBoxLayout;
  g->setLayout(hbox);
  vbox->addWidget(g);
  hbox->addWidget(eDeriv = new QComboBox);
  QStringList list;
  list << i18n("Integral") << i18n("Spline")
                                 << i18n("1st derivative")
                                 << i18n("2nd derivative")
                                 << i18n("3rd derivative")
                                 << i18n("4th derivative")
                                 << i18n("5th derivative");
  int nDeg = si->degree();
  for (int i = 0; i <= (nDeg + 1); i++)
    eDeriv->addItem(list[i]);
  int nDeriv = si->derivOrder();
  eDeriv->setCurrentIndex(nDeriv + 1);
  hbox->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding));
  hbox->addWidget(new QLabel(i18n("Lower boundary")));
  char frm = KplDoc::presentation();
  int prec = KplSettings::precision();
  hbox->addWidget(exLow = new DoubleEdit(si->xLow(), frm, prec));
  int w = exLow->fontMetrics().width(QString::number(-1e-123 / 3, frm,
                                                     prec)) + 3;
  exLow->setMinimumWidth(w);
  exLow->setEnabled(nDeriv == -1);
  connect(eDeriv, SIGNAL(activated(int)), SLOT(slotDerivChanged(int)));
  vbox->addWidget(g = new QGroupBox(i18n("Range")));
  QGridLayout* grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel("xmin"), 0, 0);
  grid->addWidget(exMin = new DoubleEdit(si->xMin(), frm, prec), 0, 1);
  exMin->setMinimumWidth(w);
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 0, 2);
  grid->addWidget(new QLabel("xmax"), 0, 3);
  grid->addWidget(exMax = new DoubleEdit(si->xMax(), frm, prec), 0, 4);
  exMax->setMinimumWidth(w);
  grid->addWidget(new QLabel(QString::fromUtf8("\316\224x")), 1, 0);
  grid->addWidget(exStep = new DoubleEdit(si->deltaX(), frm, prec), 1, 1);
  exStep->setMinimumWidth(w);
  grid->addWidget(new QLabel(i18n("x shift")), 2, 0);
  QPointF steps = m->scene()->steps(si->zValue());
  grid->addWidget(ex0 = new DoubleSpinBox(si->xOffset(), steps.x()), 2, 1);
  grid->addWidget(new QLabel(i18n("y shift")), 2, 3);
  grid->addWidget(ey0 = new DoubleSpinBox(si->yOffset(), steps.y()), 2, 4);
  grid->addWidget(new QLabel(i18n("x normalization")), 3, 0);
  grid->addWidget(efx = new DoubleEdit(si->normX()), 3, 1);
  grid->addWidget(new QLabel(i18n("y normalization")), 3, 3);
  grid->addWidget(efy = new DoubleEdit(si->normY()), 3, 4);
  vbox->addWidget(g = new QGroupBox(i18n("Representation")));
  grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel(i18n("Symbol")), 0, 0);
  grid->addWidget(symb = new SpinBox(si->symbol(), -17, 9, 1), 0, 1);
  int style = si->fillStyle();
  grid->addWidget(bSymb = new SymbolButton(symb, false, style), 0, 2);
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 0, 3);
  grid->addWidget(new QLabel(i18n("Color")), 0, 4);
  grid->addWidget(colData = new KColorButton(si->color()), 0, 5);
  colData->setMinimumWidth(50);
  grid->addWidget(new QLabel(i18n("Size")), 1, 0);
  grid->addWidget(eRelSize = new SpinBox(qRound(100 * si->relSize()), 10, 800,
                                         1), 1, 1);
  eRelSize->setSuffix(" %");
  eRelSize->setDisabled(style);
  grid->addWidget(new QLabel(i18n("Fill pattern")), 1, 4);
  grid->addWidget(fill = new BrushButton(style), 1, 5);
  vbox->addWidget(buttonBox = new QDialogButtonBox(QDialogButtonBox::Help |
                                                   QDialogButtonBox::Ok |
                                                   QDialogButtonBox::Apply |
                                                   QDialogButtonBox::Cancel));
  QPushButton* b;
  buttonBox->addButton(b = new QPushButton(i18n("&Export")),
                       QDialogButtonBox::ActionRole);
  connect(b, SIGNAL(clicked()), SLOT(slotExport()));
  buttonBox->addButton(b = new QPushButton(i18n("R&oots")),
                       QDialogButtonBox::ActionRole);
  connect(b, SIGNAL(clicked()), SLOT(slotRoots()));
  b->setEnabled(si->numKnots() && (nDeg == 3));
  resize(minimumSize());
  move(pos);
  buttonBox->button(QDialogButtonBox::Apply)->setEnabled(false);
  buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
  connect(fill, SIGNAL(changed(int)), SLOT(slotFillChanged(int)));
  connect(ex0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(ey0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(efx, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(efy, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(exMin, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(exMax, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(exStep, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(exLow, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(symb, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(eRelSize, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(colData, SIGNAL(changed(const QColor&)), SLOT(slotEdited()));
  connect(buttonBox, SIGNAL(clicked(QAbstractButton*)),
          SLOT(slotButtonClicked(QAbstractButton*)));
}

SplineDlg::~SplineDlg()
{
  delete sit;
}

void SplineDlg::apply (bool ok)
{
  buttonBox->button(QDialogButtonBox::Apply)->setEnabled(false);
  buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
  getTemp();
  SplineItem* siNew = new SplineItem(*sit);
  m->undoStack()->push(new EditCommand(siNew, si, m));
  si = siNew;
  if (ok)
    accept();
}

void SplineDlg::getTemp()
{
  sit->setDerivOrder(eDeriv->currentIndex() - 1);
  sit->setXOffset(ex0->value());
  sit->setYOffset(ey0->value());
  sit->setNormX(efx->value());
  sit->setNormY(efy->value());
  sit->setXMin(exMin->value());
  sit->setXMax(exMax->value());
  sit->setDeltaX(exStep->value());
  sit->setXLow(exLow->value());
  sit->setColor(colData->color().rgb());
  sit->setSymbol(KplItem::Symbol(symb->value()));
  sit->setRelSize(0.01 * eRelSize->value());
  sit->setFillStyle(KplItem::FillStyle(fill->value()));
}

void SplineDlg::keyPressEvent(QKeyEvent* e)
{
  if ((e->QInputEvent::modifiers() == Qt::NoModifier) &&
      (e->key() == Qt::Key_F1)) {
    buttonBox->button(QDialogButtonBox::Help)->animateClick();
    e->accept();
  } else
    QDialog::keyPressEvent(e);
}

void SplineDlg::slotApply()
{
  apply();
}

void SplineDlg::slotButtonClicked(QAbstractButton* button)
{
  int id = buttonBox->standardButton(button);
  switch (id) {
    case QDialogButtonBox::Apply:
    case QDialogButtonBox::Ok:
      apply(id == QDialogButtonBox::Ok);
      break;
    case QDialogButtonBox::Help:
      QDesktopServices::openUrl(QUrl(QStringLiteral("help:/kpl/spline.html")));
      break;
    case QDialogButtonBox::Cancel:
      reject();
  }
}

void SplineDlg::slotDerivChanged(int index)
{
  exLow->setEnabled(!index);
  slotEdited();
}

void SplineDlg::slotEdited()
{
  buttonBox->button(QDialogButtonBox::Apply)->setEnabled(true);
  buttonBox->button(QDialogButtonBox::Ok)->setEnabled(true);
}

void SplineDlg::slotExport()
{
  QUrl url;
  if (KplUtils::getWriteUrl(this, url, KplSettings::currentDir())) {
    QFile f(url.isLocalFile() ? url.path() : Utils::tempFileName());
    bool success = f.open(QIODevice::WriteOnly);
    if (success) {
      getTemp();
      QTextStream ts(&f);
      sit->exportTable(ts, KplDoc::separator(), KplDoc::presentation(),
                       KplSettings::precision());
      f.close();
      if (!url.isLocalFile())
        success = Utils::move(f.fileName(), url, this);
      if (success)
        KplUtils::setCurrentDir(url);
    }
    if (!success)
      KMessageBox::error(this, i18n("while trying to open"));
  }
}

void SplineDlg::slotFillChanged(int brush)
{
  symb->setDisabled(brush);
  bSymb->setDisabled(brush);
  eRelSize->setDisabled(brush);
  slotEdited();
}

void SplineDlg::slotRoots()
{
  RootDlg dlg(this, m, si);
  dlg.exec();
}
